#### Definition of all wrappers for 1D plotting

# Generalized lines
def axline(x=None,y=None,m=None,c=None,plabel=None,lab_loc=0,ax=None,plot_kw={},**kwargs):
	
	"""Generalised axis lines.
	
	This function aims to generalise the usage of axis lines calls (axvline/axhline) together and to allow
	lines to be specified by a slope/intercept.
	
	Parameters
	----------
	x : int or list, optional 
		x position(s) in data coordinates for a vertical line(s).
	y : int or list, optional
		y position(s) in data coordinates for a horizontal line(s).
	m : int or list, optional
		Slope(s) of diagonal axis line(s), defaults to 1 if not specified when c is given.
	c : int or list, optional
		Intercept points(s) of diagonal axis line(s), defaults to 0 if not specified when m is given.
	plabel : str, optional
		Sets label(s) for line(s) and plots legend.
	lab_loc : int, optional
		Defines the position of the legend. Defaults as lab_loc=0.
	ax : pyplot.Axes, optional
		Use the given axes to make the plot, defaults to the current axes.
	plot_kw : dict, optional
		Passes the given dictionary as a kwarg to the plotting function. Valid kwargs are Line2D properties.
	**kwargs: Line2D properties, optional
		kwargs are used to specify matplotlib specific properties such as linecolor, linewidth, antialiasing, etc.
		A list of available `Line2D` properties can be found here: 
		https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D

	Returns
	-------
	None
	"""
	
	from matplotlib.pyplot import plot, legend, gca
	from .base_func import axes_handler,plot_finalizer,dict_splicer
	
	if ax is not None:
		old_axes=axes_handler(ax)
	else:
		ax=gca()
		old_axes=ax
	
	if (not any([x,y,m,c])): # If nothing has been specified
		raise TypeError("axline() missing one of optional arguments: 'x', 'y', 'm' or 'c'")
	
	for i, val in enumerate([x,y,m,c]):
		if (val is not None):
			try: # Test whether the parameter is iterable
				temp=(k for k in val)
			except TypeError: # If not, convert to a list
				if   (i == 0): x = [x]
				elif (i == 1): y = [y]
				elif (i == 2): m = [m]
				elif (i == 3): c = [c]
	
	if (x is not None and y is not None): # Check whether both x and y were specified
		raise ValueError("'x' and 'y' cannot be both specified")
	
	if (x is not None): # Check conditions if x specified
		if (any([m,c])): # Should not specify m or c, if x given.
			raise ValueError("'{0}' cannot be specified if x specified".format('m' if m else 'c'))
		L = len(x)
	
	if (y is not None): # Check conditions if y specified
		if (any([m,c])): # Should not specify m or c, if y given.
			raise ValueError("'{0}' cannot be specified if y specified".format('m' if m else 'c'))
		L = len(y)
	
	if (m is not None):
		if (c is None): # If no intercept specified
			c = [0]*len(m) # set c to 0 for all m
		else:
			if (len(c) == 1):
				c = [c[0]]*len(m)
			elif (len(c) != len(m)):
				if (len(m) == 1):
					m = [m[0]]*len(c)
				else:
					raise ValueError("Length of c ({0}) and length of m ({1}) must be equal or otherwise 1".format(len(c),len(m)))
		L = len(m)
	elif (c is not None):
		if (m is None): # If no slope specified
			m = [1]*len(c) # set m to 1 for all c
		L = len(c)
	
	if type(plabel) is not list:
		plabel=[plabel]*L
	elif (len(plabel) != L):
		raise ValueError("Length of plabel list ({0}) must match the number of lines given ({1}).".format(len(plabel),L))
	
	# Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
	#plot_par = {**plot_kw, **kwargs} # For Python > 3.5
	plot_par = plot_kw.copy()
	plot_par.update(kwargs)
	
	# Create 'L' number of plot kwarg dictionaries to parse into each plot call
	plot_par = dict_splicer(plot_par,L,[1]*L)
	
	if (x is not None):
		for ii, xx in enumerate(x):
			ax.axvline(x=xx,**plot_par[ii],label=plabel[ii])
	if (y is not None):
		for ii, yy in enumerate(y):
			ax.axhline(y=yy,**plot_par[ii],label=plabel[ii])
	if (m is not None):
		for ii, pars in enumerate(zip(m,c)):
			mm = pars[0]; cc = pars[1]
			
			xLims = ax.get_xlim()
			yLims = ax.get_ylim()
			
			plot([xLims[0],xLims[1]],[mm*xLims[0]+cc,mm*xLims[1]+cc],label=plabel[ii],**plot_par[ii])
			
			ax.set_xlim(xLims)
			ax.set_ylim(yLims)
			
	if any(plabel):
		legend(loc=lab_loc)
	if ax is not None:
		old_axes=axes_handler(old_axes)

### Broken axis plot
def brokenplot(x,y=None,xbreak=None,ybreak=None,xlim=None,ylim=None,sep=0.05,
			   xinvert=False,yinvert=False,xlog=False,ylog=False,title=None,
			   xlabel=None,ylabel=None,plabel=None,lab_loc=0,ax=None,grid=None,plot_kw={},**kwargs):

	""" Broken Axis Plot Function
	
	Creates a standard plot call with an axis break at `xbreak` or `ybreak` for vertical or horizontal breaks.
	
	Parameters
	----------
	x : array-like or list
		If list it is assumed that each elemement is array-like. If y is not given, the given values pass to y and a
		numpy array is generated with numpy.arange() for the x values.
	y : array-like or list, optional
		If list it is assumed that each elemement is array-like.
	xbreak/ybreak : float or tuple-like, required
		The location(s) of the vertical or horizontal breaks is controlled by xbreak or ybreak, respectively.
		The value can be a single location or a tuple defining the (start, stop) points of the break. Only one 
		coordinate can be broken in a given plot.
	xlim : tuple-like, optional
		Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
	ylim : tuple-like, optional
		Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
	sep : float, optional, default: 0.05
		The separation size of the axis break, given as a fraction of the axis dimensions.
	xinvert : bool or list, optional
		If true inverts the x-axis.
	yinvert : bool or list, optional
		If true inverts the y-axis.
	xlog : bool or list, optional
		If True the scale of the x-axis is logarithmic.
	ylog : bool or list, optional
		If True the scale of the x-axis is logarithmic.
	title : str, optional
		Sets the title of the plot
	xlabel : str, optional
		Sets the label of the x-axis.
	ylabel : str, optional
		Sets the label of the y-axis.
	plabel : str, optional
		Sets the legend for the plot.
	lab_loc : int, optional
		Defines the position of the legend
	ax : pyplot.Axes, optional
		Use the given axes to make the plot, defaults to the current axes.
	grid : boolean, optional
		If not given defaults to the value defined in splotch.Params.
	plot_kw : dict, optional
		Passes the given dictionary as a kwarg to the plotting function. Valid kwargs are Line2D properties.
	**kwargs: Line2D properties, optional
		kwargs are used to specify matplotlib specific properties such as linecolor, linewidth, antialiasing, etc.
		A list of available `Line2D` properties can be found here: 
		https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D
	
	Returns
	-------
	None
	
	"""
	from .base_func import axes_handler,dict_splicer,plot_finalizer
	
	from numpy import shape, arange, ndarray
	from matplotlib.pyplot import plot, legend, show, sca
	from matplotlib.transforms import Bbox
	
	
	if ax is not None:
		old_axes=axes_handler(ax)
	if type(x) is not list or len(shape(x))==1:
		x=[x]
	L=len(x)
	if y is None:
		y=x
		x=[arange(len(x[i])) for i in range(L)]
	else:
		if type(y) is not list or len(shape(y))==1:
			y=[y]
	if type(plabel) is not list:
		plabel=[plabel]*L
	
	if type(xbreak) not in [list,tuple,ndarray]:
		xbreak = (xbreak, xbreak)
	else:
		if (len(xbreak) != 2):
			raise ValueError("xbreak must be a single value of a tuple-like list of two elements.")
	
	if (ybreak != None):
		raise NotImplementedError("ybreak not yet implemented.")
	
	# Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
	#plot_par = {**plot_kw, **kwargs} # For Python > 3.5
	plot_par = plot_kw.copy()
	plot_par.update(kwargs)
	
	# Create 'L' number of plot kwarg dictionaries to parse into each plot call
	plot_par = dict_splicer(plot_par,L,[1]*L)
	
	# Get the original axis position
	pos0 = ax.get_position(original=True)
	width0, height0 = pos0.x1 - pos0.x0, pos0.y1 - pos0.y0
	
	for i in range(L):
		ax.plot(x[i],y[i],label=plabel[i],**plot_par[i])
		
		# Get the axis limits if not already specified
		xlims = ax.get_xlim() if xlim == None else xlim
		ylims = ax.get_ylim() if ylim == None else ylim
	
		# Define the positions of the two separated axes
		if (i == 0):
			pos1 = Bbox(list(pos0.get_points()))
			pos1.x1 = pos1.x0 + (pos1.x1-pos1.x0)*(sum(xbreak)/2-xlims[0])/(xlims[1]-xlims[0]) - sep*(pos1.x1-pos1.x0)/2
			
			pos2 = Bbox(list(pos0.get_points()))
			pos2.x0 = pos2.x0 + (pos2.x1-pos2.x0)*(sum(xbreak)/2-xlims[0])/(xlims[1]-xlims[0]) + sep*(pos2.x1-pos2.x0)/2
			
			ax.set_position(pos1) # Resize the first axis
			ax2 = ax.figure.add_axes(pos2) # Add and duplicate the plotting in the second axis
			
			# Set the new axis limits at the break point
			ax.set_xlim(xlims[0],xbreak[0])
			ax2.set_xlim(xbreak[1],xlims[1])
		
		ax2.plot(x[i],y[i],label=None,**plot_par[i])
		
		
		width1, height1 = pos1.x1 - pos1.x0, pos1.y1 - pos1.y0
		width2, height2 = pos2.x1 - pos2.x0, pos2.y1 - pos2.y0
		
		dx1, dy1 = 0.01 * width0/(width0-width1-sep/2), height1*0.025
		
		dash_kw = dict(transform=ax2.transAxes, color='black', linestyle='-', marker='', clip_on=False)
		ax2.plot((0 - dx1, 0 + dx1), (0 - dy1, 0 + dy1), **dash_kw)  # bottom-right diagonal
		ax2.plot((0 - dx1, 0 + dx1), (1 - dy1, 1 + dy1), **dash_kw)  # top-right diagonal
		
		dx2, dy2 = 0.01 * width0/(width0-width2-sep/2), height2*0.025
		dash_kw.update(transform=ax.transAxes)  # switch to the left axes
		ax.plot((1 - dx2, 1 + dx2), (0 - dy2, 0 + dy2), **dash_kw)  # bottom-left sep/5iagonal
		ax.plot((1 - dx2, 1 + dx2), (1 - dy2, 1 + dy2), **dash_kw)  # top-left sep/5iagonal
		
		ax.spines['right'].set_visible(False)
		ax.tick_params(labelright=False,which='both')  # don't put tick labels at the top
		ax.yaxis.tick_left()
		
		ax2.spines['left'].set_visible(False)
		ax2.tick_params(labelleft=False,which='both')  # don't put tick labels at the top
		ax2.yaxis.tick_right()
	sca(ax)
	
	if any(plabel):
		ax.legend(loc=lab_loc)
	
	plot_finalizer(xlog,ylog,xlim,ylim,title,xlabel,ylabel,xinvert,yinvert,grid)
	
	if ax is not None:
		old_axes=axes_handler(old_axes)

def curve(expr, var=None, subs={}, permute=False, bounds=None, num=101, xlim=None, ylim=None, xinvert=False, yinvert=False,
		  xlog=False, ylog=False, title=None, xlabel=None, ylabel=None, label=True, lab_loc=0,
		  grid=None, ax=None, plot_kw={}, **kwargs):
	""" Function Plotting
	
	Plot the curve corresponding to a definingned function over the range of [from, to].
	
	Parameters
	----------
	expr : str, sympy.Expr or callable()
		An expression parsed either as a string, sympy expression or callable (function or lambda)
		which will be evaluated by the function in the range of `bounds`.
	var : str or sympy symbol, default: 'x'
		The independent variable on which to evaluate the expression (i.e. the variable
		on the x-axis). This defaults to the first non-numeric element of the expression
		or otherwise simply assumes this to be 'x'.
	subs : dict, optional
		If `expr` contains more symbols than the independent variable `var`, this dictionary
		will substitute numerical values for all additonal symbols given. `subs` is required
		if additional symbols are specified.
	bounds : float, optional
		The range over which the function will be plotted. If not given, these default to
		the current bounds of the plot.
	num : int, optional (default: 101)
		The number of values along the independent variable on which to evaulate `expr`.
	xlim : tuple-like, optional
		Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
	ylim : tuple-like, optional
		Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
	xinvert : bool or list, optional
		If true inverts the x-axis.
	yinvert : bool or list, optional
		If true inverts the y-axis.
	xlog : bool or list, optional
		If True the scale of the x-axis is logarithmic.
	ylog : bool or list, optional
		If True the scale of the x-axis is logarithmic.
	title : str, optional
		Sets the title of the plot
	xlabel : str, optional
		Sets the label of the x-axis.
	ylabel : str, optional
		Sets the label of the y-axis.
	label : bool, str or list-like, optional (default: True)
		Sets the label(s) of the curve(s) for the legend. `label` can be one of:
			- True:
				Creates a label for every curve defined by `subs`. The label will list all
				values `subs` that produced the curve. If a parameter in `subs` has only one
				value (i.e. constant amongst all curves), it will not appear in the label.
			- str:
				If a single string is given, only one label will be shown and all curves will
				be shown as the label handle.
			- list (length must equal number of curves):
				A label given to each curve that is produced.
			- False/None:
				No label will be assigned and a legend will not be shown.
	lab_loc : int, optional
		Defines the position of the legend
	grid : boolean, optional
		If not given defaults to the value defined in splotch.Params.
	ax : pyplot.Axes, optional
		Use the given axes to make the plot, defaults to the current axes.
	plot_kw : dict, optional
		Passes the given dictionary as a kwarg to the plotting function. Valid kwargs are
		Line2D properties. It is recommended that kwargs be parsed implicitly through **kwargs
		for readability.
	**kwargs: Line2D properties, optional
		kwargs are used to specify matplotlib specific properties such as linecolor, linewidth, 
		antialiasing, etc. A list of available `Line2D` properties can be found here: 
		https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D
	
	Returns
	-------
	curves : list of (or single) pyplot.Line2D object(s)
		A list of Line2D objects created for each curved create by `subs`.
	expr : Sympy.Expr
		If expr was given as a string, this returns the sympy expression created from `sympy.sympify()`.
		Otherwise, simply returns the `expr` that was given.
	
	"""
	
	from sympy import symbols, sympify, Expr
	from sympy.utilities.lambdify import lambdify
	from numpy import linspace, logspace, log10, empty, array, meshgrid, prod
	from collections import Iterable
	
	from matplotlib.pyplot import plot, legend, gca
	from matplotlib.legend_handler import HandlerPathCollection, HandlerLine2D, HandlerTuple
	from matplotlib import rcParams
	
	from .base_func import axes_handler,dict_splicer,plot_finalizer
	
	if ax is not None:
		old_axes=axes_handler(ax)
	else:
		ax=gca()
		old_axes=ax
	
	isfunc = False
	if (isinstance(expr, str)):
		expr = sympify(expr)
	elif (callable(expr)):
		isfunc = True
	elif (isinstance(expr, Expr)):
		pass
	else:
		raise TypeError(f"`expr` must be of type `str` or sympy.Expr, instead got {type(expr)}.")
	
	if (isfunc == False):
		symbs = expr.free_symbols # Get the symbols in the expression
		symbkeys = [str(sym) for sym in symbs]
		
		if (var != None): # Validate the independent variable
			if (var not in symbkeys):
				raise ValueError(f"Independent variable '{var}' was not found in 'expr'.")
		else: # Assume independent variable is 'x', otherwise, assume the first symbol.
			var = 'x' if 'x' in symbkeys or len(symbkeys)==0 else symbkeys[0]
		
		# Validate the substitution variable names
		if (subs != None):
			if (var in list(subs)):
				raise ValueError(f"Independent variable '{var}' should not be in subs.")
			for key in list(subs):
				if (key not in symbkeys):
					raise KeyError(f"Substitution variable '{key}' does not exist in 'expr'.")
	
	# The lengths of each substitute value list, len=1 if just a single value
	lens = [len(subs[key]) if (isinstance(subs[key], Iterable) and type(subs[key])!=str) else 1 for key in list(subs)]
	
	if (permute == True):
		L = prod(lens)
		perms = array(meshgrid(*subs.values())).reshape(len(subs),-1)
		permsubs = {}
		for ii, key in enumerate(list(subs)):
			permsubs[key] = perms[ii]
		subsarr = dict_splicer(permsubs,L,[1]*L)
	else:
		L = max(lens)
		subsarr = dict_splicer(subs,L,[1]*L)
	
	# Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
	#plot_par = {**plot_kw, **kwargs} # For Python > 3.5
	plot_par = plot_kw.copy()
	plot_par.update(kwargs)
	
	# Create 'L' number of plot kwarg dictionaries to parse into each plot call
	plot_par = dict_splicer(plot_par,L,[1]*L)
	
	if (bounds == None):
		if (xlim != None):
			bounds = xlim
		else:
			bounds = ax.get_xlim()
	
	
	vararr = logspace(*log10(bounds),num=num) if xlog else linspace(*bounds,num=num)
	
	curves = [None]*L
	for ii in range(L):
		if (isfunc):
			curvearr = expr(vararr, **subsarr[ii])
		else:
			func = lambdify(var, expr.subs(subsarr[ii]), 'numpy') # returns a numpy-ready function
			curvearr = func(vararr)
		
		curves[ii] = plot(vararr,curvearr,**plot_par[ii])[0]
	
	# Create the legend object
	if (label == True):
		labellist = ['']*L
		for ii in range(L): # Make a label for each of sub values
			# Create a list of the sub names and their values.
			substrs = ["{0} = {1}".format(key,subsarr[ii][key]) for jj, key in enumerate(list(subsarr[ii])) if lens[jj]>1]
			labellist[ii] = "; ".join(substrs) # join substitute strings together
		
		ax.legend(handles=curves, labels=labellist, loc=lab_loc)
	
	elif (isinstance(label, Iterable)): # A list-like iterable object or string has been given
		if (type(label) == str): # Single string given, multiple lines for the label.
			ax.legend(handles=[tuple(curves)], labels=[label], loc=lab_loc,handler_map={tuple: HandlerTuple(ndivide=L,pad=1/L**0.8)},
						handlelength=rcParams["legend.handlelength"]*L**0.8)
		else:
			if (len(label) == L): # number of labels matches number of curves
				labellist = label
			else:
				raise ValueError("Length of 'label' list does not match the number of curves defined by 'subs'.")
			ax.legend(handles=curves, labels=labellist, loc=lab_loc)
	elif (label == False or label == None): # Do not plot a legend
		pass
	else:
		raise ValueError("Value of 'label' is invalid. Expected bool, str, iterable or None, but got '{0}'.".format(label))
	
	plot_finalizer(xlog,ylog,xlim,ylim,title,xlabel,ylabel,xinvert,yinvert,grid)
	
	if ax is not None:
		old_axes=axes_handler(old_axes)
	
	return(curves[0] if len(curves)==1 else curves, expr)

#Histogram
def hist(data,bin_type=None,bins=None,dens=True,cumul=None,scale=None,hist_type=None,v=None,vstat=None,xlim=None,ylim=None,
			xinvert=False,yinvert=False,xlog=False,ylog=None,title=None,xlabel=None,ylabel=None,plabel=None,lab_loc=0,
			ax=None,grid=None,plot_kw={},output=None,**kwargs):
	
	"""1D histogram function.
	
	The plotting is done with pyplot.plot(), so histograms are shown with interpolated curves instead of the
	more common stepwise curve. For this reason splotch.histstep is a better choice for small datasets. 
	
	Parameters
	----------
	data : array-like or list
		If list it is assumed that each elemement is array-like.
	bin_type : {'number','width','edges','equal'}, optional
		Defines how is understood the value given in bins: 'number' for the desired number of bins, 'width' for the width
		of the bins, 'edges' for the edges of bins, and 'equal' for making bins with equal number of elements (or as close
		as possible). If not given it is inferred from the data type of bins: 'number' if int, 'width' if float and 'edges'
		if ndarray.
	bins : int, float, array-like or list, optional
		Gives the values for the bins, according to bin_type.
	dens :  bool or list, optional
		If false the histogram returns raw counts.
	cumul : bool or list, optional
		If true, produces a cumulative distribution instead of a histogram.
	scale : float or list, optional
		Scaling the counts.
	hist_type : str, optional.
		Defines the type of histogram to be drawn. 'smooth' and 'step' produce lines, with the former drawing lines conecting
		the values of each bin positioned on their centre, and the latter drawing a stepwise line, with the edges of each step
		coinciding with the bin edges. 'bar' produces a bar plot. All have filled version (i.e., 'smoothfilled'), which fills
		the space between the edges of the histogram and 0.
	v : array-like or list, optional
		If a valid argument is given in cstat, defines the value used for the binned statistics.
	vstat : str, function  or list, optional
		Must be or contain one of the valid str arguments for the statistics variable in scipy.stats.binned_statistic
		('mean’, 'median’, 'count’, 'sum’, 'min’ or 'max’) or function(s) that takes a 1D array and outputs an integer
		 or float.
	xlim : tuple-like, optional
		Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
	ylim : tuple-like, optional
	xinvert : bool or list, optional
		If true inverts the x-axis.
		Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
	yinvert : bool or list, optional
		If true inverts the y-axis.
	xlog : bool or list, optional
		If True the scale of the x-axis is logarithmic.
	ylog : bool or list, optional
		If True the scale of the x-axis is logarithmic.
	title : str, optional
		Sets the title of the plot
	xlabel : str, optional
		Sets the label of the x-axis.
	ylabel : str, optional
		Sets the label of the y-axis.
	plabel : str, optional
		Sets the legend for the plot.
	lab_loc : int, optional
		Defines the position of the legend
	ax : pyplot.Axes, optional
		Use the given axes to make the plot, defaults to the current axes.
	grid : boolean, optional
		If not given defaults to the value defined in splotch.Params.
	plot_par : dict, optional
		Passes the given dictionary as a kwarg to the plotting function.
	output : boolean, optional
		If True, returns the edges and values of the histogram.
	
	Returns
	-------
	n : list
		List containing the arrays with the values for each histogram drawn. Only provided if output is True.
	bins_edges : list
		List containing the arrays with the bin edges for each of the histograms drawn. Only provided if output is True.
	"""
	
	from numpy import cumsum as np_cumsum, max as np_max, sum as np_sum
	from scipy.stats import binned_statistic
	from numpy import array, diff, histogram, inf, nan, ones, where
	from matplotlib.pyplot import bar, fill_between, gca, legend, plot, rcParams, step
	from .base_func import axes_handler,bin_axis,dict_splicer,plot_finalizer,step_filler
	
	if ax is not None:
		old_axes=axes_handler(ax)
	if type(data) is not list:
		data=[data]
	L=len(data)
	if type(bin_type) is not list:
		bin_type=[bin_type]*L
	if type(bins) is not list:
		if bins is not None:
			bins=[bins]*L
		else:
			bins=[int((len(d))**0.4) for d in data]
	if type(dens) is not list:
		dens=[dens]*L
	if type(cumul) is not list:
		cumul=[cumul]*L
	if type(scale) is not list:
		scale=[scale]*L
	if type(v) is not list:
		v=[v]*L
	if type(vstat) is not list:
		vstat=[vstat]*L
	if type(plabel) is not list:
		plabel=[plabel]*L
	if None in [ylog,hist_type,output]:
		from .defaults import Params
		if ylog is None:
			ylog=Params.hist1D_yaxis_log
		if hist_type is None:
			hist_type=Params.hist1D_histtype
		if output is None:
			output=Params.hist1D_output
	if type(hist_type) is not list:
		hist_type=[hist_type]*L
	
	# Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
	#plot_par = {**plot_kw, **kwargs} # For Python > 3.5
	plot_par = plot_kw.copy()
	plot_par.update(kwargs)
	# Check if width is given as a kwarg
	if 'width' in plot_par.keys():
		import warnings
		warnings.warn('Received kwarg width, this will be ignored in the histogram',UserWarning)
		if hist_type!='bar':
			temp=plot_par.pop('width')
	# Create 'L' number of plot kwarg dictionaries to parse into each plot call
	plot_par = dict_splicer(plot_par,L,[1]*L)
	
	plot_type={'smooth':plot,'smoothfilled':fill_between,'step':step,'stepfilled':step_filler,'bar':bar,'barfilled':bar}
	hist_centre={'smooth':True,'smoothfilled':True,'step':False,'stepfilled':False,'bar':False,'barfilled':False}
	bin_edges=[]
	n_return=[]
	
	for i in range(L):
		temp_data,bins_hist,bins_plot=bin_axis(data[i],bin_type[i],bins[i],log=xlog,plot_centre=hist_centre[hist_type[i]])
		if vstat[i]:
			temp_y=binned_statistic(temp_data,v[i],statistic=vstat[i],bins=bins_hist)[0]
		else:
			temp_y=histogram(temp_data,bins=bins_hist,density=dens[i])[0]
		if cumul[i]:
			temp_y=np_cumsum(temp_y)
			if dens[i]:
				temp_y/=np_max(temp_y)
		if scale[i]:
			if dens[i]:
				temp_y*=len(data[i])/scale[i]
			else:
				temp_y/=scale[i]
		if ylog:
			temp_y=where(temp_y==0,nan,temp_y)
		y=temp_y
		if hist_type[i]=='step':
			y=array([y[0]]+[j for j in y])
		if 'bar' in hist_type[i]:
			#prop_cycle=rcParams['axes.prop_cycle']
			#barcolor=prop_cycle.by_key()['color']
			plot_par[i]['width']=diff(bins_plot)
			bins_plot=(bins_plot[0:-1]+bins_plot[1:])/2
			if hist_type[i]=='bar':
				if 'edgecolor' not in plot_par[i].keys():
					p=plot(bins_plot[0],0)
					plot_par[i]['edgecolor']=p[0].get_color()
					p.pop()
					temp_ax=gca()
					temp_ax.relim()
					temp_ax.autoscale()
				plot_par[i]['fill']=False
		plot_type[hist_type[i]](bins_plot,y,label=plabel[i],**plot_par[i])
		bin_edges.append(bins_plot)
		n_return.append(temp_y)
	if any(plabel):
		legend(loc=lab_loc)
	plot_finalizer(xlog,ylog,xlim,ylim,title,xlabel,ylabel,xinvert,yinvert,grid)
	if ax is not None:
		old_axes=axes_handler(old_axes)
	if output:
		return(n_return,bin_edges)

#Step histogram
def hist_legacy(data,bin_num=None,dens=True,xlim=None,ylim=None,xinvert=False,yinvert=False,xlog=False,ylog=True,
			title=None,xlabel=None,ylabel=None,plabel=None,lab_loc=0,ax=None,grid=None,plot_kw={}):
	
	"""'Clasic' 1D histogram function.
	
	This function is designed around pyplot.hist(), so it lacks the functionality to use arbitraty y-axis
	normalisation of splotch.hist().
	It is better choice for small datasets, as it plots with stepwise curves, instead of the interpolated
	ones of splotch.hist().
	
	Parameters
	----------
	data : array-like or list
		If list it is assumed that each elemement is array-like.
	bin_num : int or list, optional
		Number of bins.
	dens :  bool or list, optional
		If false the histogram returns raw counts.
	xlim : tuple-like, optional
		Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
	ylim : tuple-like, optional
		Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
	xinvert : bool or list, optional
		If true inverts the x-axis.
	yinvert : bool or list, optional
		If true inverts the y-axis.
	xlog : bool or list, optional
		If True the scale of the x-axis is logarithmic.
	ylog : bool or list, optional
		If True the scale of the x-axis is logarithmic.
	title : str, optional
		Sets the title of the plot
	xlabel : str, optional
		Sets the label of the x-axis.
	ylabel : str, optional
		Sets the label of the y-axis.
	plabel : str, optional
		Sets the legend for the plot.
	lab_loc : int, optional
		Defines the position of the legend
	ax : pyplot.Axes, optional
		Use the given axes to make the plot, defaults to the current axes.
	grid : boolean, optional
		If not given defaults to the value defined in splotch.Params.
	plot_par : dict, optional
		Passes the given dictionary as a kwarg to the plotting function.
	
	Returns
	-------
	None
	"""
	
	from matplotlib.pyplot import hist, legend
	from .base_func import axes_handler,dict_splicer,plot_finalizer
	
	if ax is not None:
		old_axes=axes_handler(ax)
	if type(data) is not list:
		data=[data]
	L=len(data)
	if bin_num is None:
		bin_num=[int((len(d))**0.4) for d in data]
	if type(bin_num) is not list:
		bin_num=[bin_num+1]*L
	if type(plabel) is not list:
		plabel=[plabel]*L
	plot_par=dict_splicer(plot_kw,L,[len(x) for x in data])
	for i in range(L):
		temp_data,bins,temp=bin_axis(data[i],bin_num[i],log=xlog)
		hist(temp_data,bins=bins,density=dens,label=plabel[i],**plot_par[i])
	if any(plabel):
		legend(loc=lab_loc)
	plot_finalizer(xlog,ylog,xlim,ylim,title,xlabel,ylabel,xinvert,yinvert,grid)
	if ax is not None:
		old_axes=axes_handler(old_axes)

#Plots
def plot(x,y=None,xlim=None,ylim=None,xinvert=False,yinvert=False,xlog=False,ylog=False,title=None,xlabel=None,
			ylabel=None,plabel=None,lab_loc=0,ax=None,grid=None,plot_kw={},**kwargs):
	
	"""Base plotting function.
	
	This is a wrapper for pyplot.plot().
	
	Parameters
	----------
	x : array-like or list
		If list it is assumed that each elemement is array-like. If y is not given, the given values pass to y and a
		numpy array is generated with numpy.arange() for the x values.
	y : array-like or list, optional
		If list it is assumed that each elemement is array-like.
	xlim : tuple-like, optional
		Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
	ylim : tuple-like, optional
		Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
	xinvert : bool or list, optional
		If true inverts the x-axis.
	yinvert : bool or list, optional
		If true inverts the y-axis.
	xlog : bool or list, optional
		If True the scale of the x-axis is logarithmic.
	ylog : bool or list, optional
		If True the scale of the x-axis is logarithmic.
	title : str, optional
		Sets the title of the plot
	xlabel : str, optional
		Sets the label of the x-axis.
	ylabel : str, optional
		Sets the label of the y-axis.
	plabel : str, optional
		Sets the legend for the plot.
	lab_loc : int, optional
		Defines the position of the legend
	ax : pyplot.Axes, optional
		Use the given axes to make the plot, defaults to the current axes.
	grid : boolean, optional
		If not given defaults to the value defined in splotch.Params.
	plot_kw : dict, optional
		Passes the given dictionary as a kwarg to the plotting function. Valid kwargs are Line2D properties.
	**kwargs: Line2D properties, optional
		kwargs are used to specify matplotlib specific properties such as linecolor, linewidth, antialiasing, etc.
		A list of available `Line2D` properties can be found here: 
		https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D
	
	Returns
	-------
	None
	"""

	from numpy import shape, arange
	from matplotlib.pyplot import plot, legend
	from .base_func import axes_handler,dict_splicer,plot_finalizer
	
	
	if ax is not None:
		old_axes=axes_handler(ax)
	if type(x) is not list or len(shape(x))==1:
		x=[x]
	L=len(x)
	if y is None:
		y=x
		x=[arange(len(x[i])) for i in range(L)]
	else:
		if type(y) is not list or len(shape(y))==1:
			y=[y]
	if type(plabel) is not list:
		plabel=[plabel]*L
	
	# Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
	#plot_par = {**plot_kw, **kwargs} # For Python > 3.5
	plot_par = plot_kw.copy()
	plot_par.update(kwargs)
	
	# Create 'L' number of plot kwarg dictionaries to parse into each plot call
	plot_par = dict_splicer(plot_par,L,[1]*L)
	
	for i in range(L):
		plot(x[i],y[i],label=plabel[i],**plot_par[i])
	if any(plabel):
		legend(loc=lab_loc)
	plot_finalizer(xlog,ylog,xlim,ylim,title,xlabel,ylabel,xinvert,yinvert,grid)
	if ax is not None:
		old_axes=axes_handler(old_axes)




#Object containing the default values for SPLOTCH functions
class Params:
	#Contours
	cont_filled=False
	sigcont_percent=[68.27,95.45]
	sigcont_cmap='viridis'
	sigcont_clim=[0.33,0.67]
	sigcont_linestyle=['solid','dashed','dotted']
	sigcont_output=False
	#Grid
	grid=False
	grid_which='major'
	grid_axis='both'
	#Histograms
	hist1D_output=False
	hist1D_histtype='step'
	hist1D_yaxis_log=False
	hist2D_caxis_log=False
	hist2D_output=False
	#Images
	img_caxis_log=False

class Params0():
	def __init__(self):
		self._params = {}
	def __setitem__(self, key, val):
		self._params[key] = val
	def __getitem__(self, key):
		return self._params[key]

#params = Params0()
#params['test'] = 'yay!'



#### Definition of all wrappers for 2D plotting

#Level contours
def cont(z,x=None,y=None,filled=None,xlim=None,ylim=None,xinvert=False,yinvert=False,xlog=False,ylog=False,title=None,xlabel=None,
			ylabel=None,lab_loc=0,ax=None,grid=None,plot_kw={},**kwargs):
	
	"""Level contour plotting function (legacy name).
	
	This provides legacy compatibily for old code using the original name of the function.
	This function will eventually be removed, so consider switching to plots_2d.contour().
	"""
	
	import warnings
	
	warnings.warn('This function provides legacy support and it will be removed in the future. Use plots_2d.contour() instead.',FutureWarning)
	
	contour(z,x,y,filled,xlim,ylim,xinvert,yinvert,xlog,ylog,title,xlabel,ylabel,lab_loc,ax,grid,plot_kw={},**kwargs)

def contour(z,x=None,y=None,filled=None,xlim=None,ylim=None,xinvert=False,yinvert=False,xlog=False,ylog=False,title=None,xlabel=None,
			ylabel=None,lab_loc=0,ax=None,grid=None,plot_kw={},**kwargs):
	
	"""Level contour plotting function.
	
	This is a wrapper for pyplot.contour() and pyplot.contourf().
	
	Parameters
	----------
	z : array-like
		The height values to draw the contours.
	x : array-like, optional
		Position of data points in the x axis.
	y : array-like, optional
		Position of data points in the y axis.
	filled: boolean, optional
		If True draws filled contours. If not given defaults to the value defined in splotch.Params.
	xlim : tuple-like, optional
		Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
	ylim : tuple-like, optional
		Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
	xinvert : bool, optional
		If true inverts the x-axis.
	yinvert : bool, optional
		If true inverts the y-axis.
	xlog : bool, optional
		If True the scale of the x-axis is logarithmic. If not given defaults to the value defined in splotch.Params.
	ylog : bool, optional
		If True the scale of the x-axis is logarithmic. If not given defaults to the value defined in splotch.Params.
	title : str, optional
		Sets the title of the plot
	xlabel : str, optional
		Sets the label of the x-axis.
	ylabel : str, optional
		Sets the label of the y-axis.
	lab_loc : int, optional
		Defines the position of the legend
	ax : pyplot.Axes, optional
		Use the given axes to make the plot, defaults to the current axes.
	grid : boolean, optional
		If not given defaults to the value defined in splotch.Params.
	output : boolean, optional
		If True, returns the edges and values of the underlying histogram plus the levels of the contours.
	plot_kw : dict, optional
		Passes the given dictionary as a kwarg to the plotting function. Valid kwargs are QuadContourSet properties.
	**kwargs: QuadContourSet properties, optional
		kwargs are used to specify matplotlib specific properties such as cmap, linewidths, hatches, etc.
		The list of available properties can be found here: 
		https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.contour.html
	
	Returns
	-------
	bin_edges_x : array
		The bin edges for the x axis.
	bin_edges_y : array
		The bin edges for the y axis.
	n : array
		The values of the underlying histogram.
	l : array
		The levels for the contours.
	"""
	
	from numpy import shape, linspace
	from matplotlib.pyplot import contour,contourf, legend
	from .base_func import axes_handler,dict_splicer,plot_finalizer
	
	if ax is not None:
		old_axes=axes_handler(ax)
	if filled is None:
		from .defaults import Params
		filled=Params.cont_filled
	if x is None:
		x=linspace(0,1,z.shape[0])
	if y is None:
		y=linspace(0,1,z.shape[1])
	
	# Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
	#plot_par = {**plot_kw, **kwargs} # For Python > 3.5
	plot_par = plot_kw.copy()
	plot_par.update(kwargs)
	
	# Create 'L' number of plot kwarg dictionaries to parse into each plot call
	
	plotf={False:contour,True:contourf}
	plotf[filled](x,y,z,**plot_par)
	
	plot_finalizer(xlog,ylog,xlim,ylim,title,xlabel,ylabel,xinvert,yinvert,grid)
	if ax is not None:
		old_axes=axes_handler(old_axes)

#Errorbars
def errbar(x,y,xerr=None,yerr=None,xlim=None,ylim=None,xinvert=False,yinvert=False,xlog=False,ylog=False,
	title=None,xlabel=None,ylabel=None,plabel=None,lab_loc=0,ax=None,grid=None,plot_kw={},**kwargs):
	
	"""Errorbar plotting function (legacy name).
	
	This provides legacy compatibily for old code using the original name of the function.
	This function will eventually be removed, so consider switching to plots_2d.errorbar().
	"""
	
	import warnings
	
	warnings.warn('This function provides legacy support and it will be removed in the future. Use plots_2d.errorbar() instead.',FutureWarning)
	
	errorbar(x,y,xerr,yerr,xlim,ylim,xinvert,yinvert,xlog,ylog,title,xlabel,ylabel,plabel,lab_loc,ax,grid,plot_kw,**kwargs)

def errorbar(x,y,xerr=None,yerr=None,xlim=None,ylim=None,xinvert=False,yinvert=False,xlog=False,ylog=False,
				title=None,xlabel=None,ylabel=None,plabel=None,lab_loc=0,ax=None,grid=None,plot_kw={},**kwargs):
	
	"""Errorbar plotting function.
	
	This is a wrapper for pyplot.errorbar().
	
	Parameters
	----------
	x : array-like or list
		If list it is assumed that each elemement is array-like.
	y : array-like or list
		If list it is assumed that each elemement is array-like.
	xerr : array-like or list, optional
		Defines the length of the errobars in the x-axis. If list it is assumed that each elemement is array-like.
	yerr : array-like or list, optional
		Defines the length of the errobars in the y-axis. If list it is assumed that each elemement is array-like.
	xlim : tuple-like, optional
		Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
	ylim : tuple-like, optional
		Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
	xinvert : bool or list, optional
		If true inverts the x-axis.
	yinvert : bool or list, optional
		If true inverts the y-axis.
	xlog : bool or list, optional
		If True the scale of the x-axis is logarithmic.
	ylog : bool or list, optional
		If True the scale of the x-axis is logarithmic.
	title : str, optional
		Sets the title of the plot
	xlabel : str, optional
		Sets the label of the x-axis.
	ylabel : str, optional
		Sets the label of the y-axis.
	plabel : str, optional
		Sets the legend for the plot.
	lab_loc : int, optional
		Defines the position of the legend
	ax : pyplot.Axes, optional
		Use the given axes to make the plot, defaults to the current axes.
	grid : boolean, optional
		If not given defaults to the value defined in splotch.Params.
	plot_kw : dict, optional
		Passes the given dictionary as a kwarg to the plotting function. Valid kwargs are Line2D properties.
	**kwargs: Line2D properties, optional
		kwargs are used to specify matplotlib specific properties such as linecolor, linewidth, antialiasing, etc.
		A list of available `Line2D` properties can be found here: 
		https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D
	
	Returns
	-------
	None
	"""
	
	from matplotlib.pyplot import errorbar, legend
	from .base_func import axes_handler,dict_splicer,plot_finalizer
	
	if ax is not None:
		old_axes=axes_handler(ax)
	if type(x) is not list:
		x=[x]
	if type(y) is not list:
		y=[y]
	if type(xerr) is not list:
		xerr=[xerr]
	if type(yerr) is not list:
		yerr=[yerr]
	L=len(x)
	if type(plabel) is not list:
		plabel=[plabel]*L

	# Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
	#plot_par = {**plot_kw, **kwargs} # For Python > 3.5
	plot_par = plot_kw.copy()
	plot_par.update(kwargs)

	# Create 'L' number of plot kwarg dictionaries to parse into each plot call
	plot_par = dict_splicer(plot_par,L,[1]*L)

	for i in range(L):
		errorbar(x[i],y[i],xerr=xerr[i],yerr=yerr[i],label=plabel[i],**plot_par[i])
	if any(plabel):
		legend(loc=lab_loc)
	plot_finalizer(xlog,ylog,xlim,ylim,title,xlabel,ylabel,xinvert,yinvert,grid)
	if ax is not None:
		old_axes=axes_handler(old_axes)

#Errorboxes
def errbox(x,y,xerr=None,yerr=None,xlim=None,ylim=None,xinvert=False,yinvert=False,xlog=False,ylog=False,boxtype='ellipse',
	title=None,xlabel=None,ylabel=None,plabel=None,grid=None,lab_loc=0,ax=None,plot_kw={},**kwargs):
	
	"""Errorbox plotting function (legacy name).
	
	This provides legacy compatibily for old code using the original name of the function.
	This function will eventually be removed, so consider switching to plots_2d.errorbox().
	"""
	
	import warnings
	
	warnings.warn('This function provides legacy support and it will be removed in the future. Use plots_2d.errorbox() instead.',FutureWarning)
	
	errorbox(x,y,xerr,yerr,xlim,ylim,xinvert,yinvert,xlog,ylog,boxtype,title,xlabel,ylabel,plabel,grid,lab_loc,ax,plot_kw,**kwargs)

def errorbox(x,y,xerr=None,yerr=None,xlim=None,ylim=None,xinvert=False,yinvert=False,xlog=False,ylog=False,boxtype='ellipse',
			title=None,xlabel=None,ylabel=None,plabel=None,grid=None,lab_loc=0,ax=None,plot_kw={},**kwargs):
	
	"""Errorbox plotting function.
	
	This is a wrapper around matplotlib PatchCollections with a matplotlib errorbar functionality.
	
	Parameters
	----------
	x : array-like or list
		If list it is assumed that each elemement is array-like.
	y : array-like or list
		If list it is assumed that each elemement is array-like.
	xerr : array-like or list, optional
		Defines the length of the errobars in the x-axis. If list it is assumed that each elemement is array-like.
	yerr : array-like or list, optional
		Defines the length of the errobars in the y-axis. If list it is assumed that each elemement is array-like.
	xlim : tuple-like, optional
		Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
	ylim : tuple-like, optional
		Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
	xinvert : bool or list, optional
		If true inverts the x-axis.
	yinvert : bool or list, optional
		If true inverts the y-axis.
	xlog : bool or list, optional
		If True the scale of the x-axis is logarithmic.
	ylog : bool or list, optional
		If True the scale of the x-axis is logarithmic.
	boxtype : str
		The type of box to plot, patch types include: ellipse | rectangle (Default: ellipse).
	title : str, optional
		Sets the title of the plot
	xlabel : str, optional
		Sets the label of the x-axis.
	ylabel : str, optional
		Sets the label of the y-axis.
	plabel : str, optional
		Sets the legend for the plot.
	lab_loc : int, optional
		Defines the position of the legend
	ax : pyplot.Axes, optional
		Use the given axes to make the plot, defaults to the current axes.
	grid : boolean, optional
		If not given defaults to the value defined in splotch.Params.
	plot_kw : dict, optional
		Passes the given dictionary as a kwarg to the plotting function. Valid kwargs are Patches properties.
	**kwargs: Patch properties, optional
		kwargs are used to specify matplotlib specific properties such as facecolor, linestyle, alpha, etc.
		A list of available `Patch` properties can be found here: 
		https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.patches.Rectangle.html

	
	Returns
	-------
	None
	"""
	
	from matplotlib.pyplot import errorbar, legend
	from .base_func import axes_handler,dict_splicer,plot_finalizer

	from numpy import shape, full, array

	from matplotlib.collections import PatchCollection
	from matplotlib.patches import Ellipse, Rectangle
	
	if ax is not None:
		old_axes=axes_handler(ax)
	if type(x) is not list:
		x=[x]
	if type(y) is not list:
		y=[y]
	if type(xerr) is not list:
		xerr=[xerr]
	if type(yerr) is not list:
		yerr=[yerr]

	L=len(x)
	if type(plabel) is not list:
		plabel=[plabel]*L

	# Validate format of xerr and yerr
	for i in range(L):
		# x-axis errors
		if (shape(xerr[i]) == ()): # single error for all points
			xerr[i] = full((2,len(x[i])), xerr[i])
		else:
			if (len(shape(xerr[i])) == 1):
				if (shape(xerr[i])[0] == len(x[i])): # single error for each point
					xerr[i] = array([xerr[i], xerr[i]])
				elif (shape(xerr[i])[0] == 2): # separate upper and lower errors for all points
					xerr[i] = full((len(x[i]), 2), xerr[i]).T
				else:
					print('ding') # Raise exception for invalid length of points
			elif (len(shape(xerr[i])) == 2): # separate upper and lower errors for each point
				xerr[i] = array(xerr[i])
				if (shape(xerr[i])[0] != 2 or shape(xerr[i])[1] != len(x[i])):
					print('dong') # Raise exception for invalid length of points

		# y-axis errors
		if (shape(yerr[i]) == ()): # single error for all points
			yerr[i] = full((2,len(y[i])), yerr[i])
		else:
			if (len(shape(yerr[i])) == 1):
				if (shape(yerr[i])[0] == len(y[i])): # single error for each point
					yerr[i] = array([yerr[i], yerr[i]])
				elif (shape(yerr[i])[0] == 2): # separate upper and lower errors for all points
					yerr[i] = full((len(y[i]), 2), yerr[i]).T
				else:
					print('ding') # Raise exception for invalid length of points
			elif (len(shape(yerr[i])) == 2): # separate upper and lower errors for each point
				yerr[i] = array(yerr[i])
				if (shape(yerr[i])[0] != 2 or shape(yerr[i])[1] != len(y[i])):
					print('dong') # Raise exception for invalid length of points


	# Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
	#plot_par = {**plot_kw, **kwargs} # For Python > 3.5
	plot_par = plot_kw.copy()
	plot_par.update(kwargs)

	# Create 'L' number of plot kwarg dictionaries to parse into each plot call
	plot_par = dict_splicer(plot_par,L,[1]*L)
 
	PathColls = []
	# Loop over data points; create box/ellipse from errors at each point
	for i in range(L):
		errorboxes = []
		for xx, yy, xe, ye in zip(x[i], y[i], xerr[i].T, yerr[i].T):
			if (boxtype.lower().startswith('rect')):
				errorboxes.append( Rectangle((xx - xe[0], yy - ye[0]), xe.sum(), ye.sum()) )
			elif (boxtype.lower().startswith('ell')):
				errorboxes.append( Ellipse((xx - xe[0], yy - ye[0]), xe.sum(), ye.sum()) )
			else:
				print('dang')

		# Create and add patch collection with specified colour/alpha
		pc = PatchCollection(errorboxes, **plot_par[i])
		ax.add_collection(pc)

	# for i in range(L):
	# 	errorbar(x[i],y[i],xerr=xerr[i],yerr=yerr[i],label=plabel[i],**plot_par[i])
	if any(plabel):
		legend(loc=lab_loc)
	plot_finalizer(xlog,ylog,xlim,ylim,title,xlabel,ylabel,xinvert,yinvert,grid)
	if ax is not None:
		old_axes=axes_handler(old_axes)


# Histogram and 2D binned statistics
def hist2D(x,y,bin_type=None,bins=None,dens=True,scale=None,c=None,cstat=None,xlim=None,ylim=None,clim=[None,None],nmin=0, 
			xinvert=False,yinvert=False,cbar_invert=False,xlog=False,ylog=False,clog=None,title=None,xlabel=None,
			ylabel=None,clabel=None,lab_loc=0,ax=None,grid=None,output=None,plot_kw={},**kwargs):
	
	"""2D histogram function.
	
	Parameters
	----------
	x : array-like
		Position of data points in the x axis.
	y : array-like
		Position of data points in the y axis.
	bin_type : {'number','width','edges','equal'}, optional
		Defines how is understood the value given in bins: 'number' for givinf the desired number of bins, 'width' for
		the width of the bins, 'edges' for the edges of bins, and 'equal' for making bins with equal number of elements
		(or as close as possible). If not given it is inferred from the data type of bins: 'number' if int, 'width' if
		float and 'edges' if ndarray.
	bins : int, float, array-like or list, optional
		Gives the values for the bins, according to bin_type.
	dens : bool or list, optional
		If false the histogram returns raw counts.
	scale : float or list, optional
		Scaling of the data counts.
	c : array-like, optional
		If a valid argument is given in cstat, defines the value used for the binned statistics.
	cstat : str or function, optional
		Must be one of the valid str arguments for the statistics variable in scipy.stats.binned_statistic_2d
		('mean’, 'median’, 'count’, 'sum’, 'min’ or 'max’) or a function that takes a 1D array and outputs an integer
		 or float.
	xlim : tuple-like, optional
		Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
	ylim : tuple-like, optional
		Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
	clim : list, optional
		Defines the limits of the colour map ranges, it must contain two elements (lower and higer limits).
	nmin : int, optional (default: 0)
		The minimum number of points required in a bin in order to be plotted.
	xinvert : bool, optional
		If true inverts the x-axis.
	yinvert : bool, optional
		If true inverts the y-axis.
	cbar_invert : bool, optional
		If True inverts the direction of the colour bar (not the colour map).
	xlog : bool, optional
		If True the scale of the x-axis is logarithmic.
	ylog : bool, optional
		If True the scale of the x-axis is logarithmic.
	clog : bool, optional
		If True, the colour map is changed from linear to logarithmic.
	title : str, optional
		Sets the title of the plot
	xlabel : str, optional
		Sets the label of the x-axis.
	ylabel : str, optional
		Sets the label of the y-axis.
	clabel : str, optional
		Sets the legend for the colour axis.
	lab_loc : int, optional
		Defines the position of the legend
	ax : pyplot.Axes, optional
		Use the given axes to make the plot, defaults to the current axes.
	grid : boolean, optional
		If not given defaults to the value defined in splotch.Params.
	output : boolean, optional
		If True, returns the edges and values of the histogram.
	plot_kw : dict, optional
		Explicit dictionary of kwargs to be parsed to matplotlib scatter function.
		Parameters will be overwritten if also given implicitly as a **kwarg.
	**kwargs : pcolormesh properties, optional
		kwargs are used to specify matplotlib specific properties such as cmap, norm, edgecolors etc.
		https://matplotlib.org/api/_as_gen/matplotlib.pyplot.pcolormesh.html

	Returns
	-------
	n : array
		The values of the histogram. Only provided if output is True.
	bin_edges_x : array
		The bin edges for the x axis. Only provided if output is True.
	bin_edges_y : array
		The bin edges for the y axis. Only provided if output is True.
	"""
	
	from numpy import nan
	from matplotlib.colors import LogNorm
	from matplotlib.pyplot import pcolormesh, colorbar
	from .base_func import axes_handler,basehist2D,plot_finalizer
	
	if ax is not None:
		old_axes=axes_handler(ax)
	if type(bin_type) is not list:
		bin_type=[bin_type]*2
	if type(bins) not in [list,tuple]:
		if bins is None:
			bins=int((len(x))**0.4)
		bins=[bins]*2
	X,Y,Z=basehist2D(x,y,c,bin_type,bins,scale,dens,cstat,xlog,ylog)
	_,_,counts = basehist2D(x,y,c,bin_type,bins,scale,dens,'count',xlog,ylog) # Also get counts for number threshold cut

	# Cut bins which do not meet the number count threshold
	Z[counts<nmin] = nan
	
	# Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
	#plot_par = {**plot_kw, **kwargs} # For Python > 3.5
	plot_par = plot_kw.copy()
	plot_par.update(kwargs)
	
	if None in (clog,output):
		from .defaults import Params
		if clog is None:
			clog=Params.hist2D_caxis_log
		if output is None:
			output=Params.hist2D_output
	if clog:
		pcolormesh(X,Y,Z.T,norm=LogNorm(vmin=clim[0],vmax=clim[1],clip=True),**plot_par)
	else:
		if cstat is None:
			Z[Z==0]=nan
		pcolormesh(X,Y,Z.T,vmin=clim[0],vmax=clim[1],**plot_par)
	if clabel is not None:
		cbar=colorbar()
		cbar.set_label(clabel)
		if cbar_invert:
			cbar.ax.invert_yaxis()
	plot_finalizer(xlog,ylog,xlim,ylim,title,xlabel,ylabel,xinvert,yinvert,grid)
	if ax is not None:
		old_axes=axes_handler(old_axes)
	if output:
		return(Z.T,X,Y)

# Image
def img(im,x=None,y=None,xlim=None,ylim=None,clim=[None,None],cmin=0,xinvert=False,yinvert=False,cbar_invert=False,clog=None,
		title=None,xlabel=None,ylabel=None,clabel=None,lab_loc=0,ax=None,grid=None,plot_kw={},**kwargs):
	
	"""2D pixel-based image plotting function.
	
	Parameters
	----------
	im : array-like
		Value for each pixel in an x-y 2D array, where the first dimension is the x-position and the second is
		the y-position.
	x : array-like, optional
		Position of data points in the x axis.
	y : array-like, optional
		Position of data points in the y axis.
	xlim : tuple-like, optional
		Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
	ylim : tuple-like, optional
		Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
	clim : list, optional
		Defines the limits of the colour map ranges, it must contain two elements (lower and higer limits).
	clog : bool, optional
		If True, the colour map is changed from linear to logarithmic.
	xinvert : bool, optional
		If true inverts the x-axis.
	yinvert : bool, optional
		If true inverts the y-axis.
	cbar_invert : bool, optional
		If True inverts the direction of the colour bar (not the colour map).
	title : str, optional
		Sets the title of the plot
	xlabel : str, optional
		Sets the label of the x-axis.
	ylabel : str, optional
		Sets the label of the y-axis.
	clabel : str, optional
		Sets the legend for the colour axis.
	lab_loc : int, optional
		Defines the position of the legend
	ax : pyplot.Axes, optional
		Use the given axes to make the plot, defaults to the current axes.
	grid : boolean, optional
		If not given defaults to the value defined in splotch.Params.
	plot_kw : dict, optional
		Explicit dictionary of kwargs to be parsed to matplotlib pcolormesh function.
		Parameters will be overwritten if also given implicitly as a **kwarg.
	**kwargs : pcolormesh properties, optional
		kwargs are used to specify matplotlib specific properties such as `cmap`, `marker`, `norm`, etc.
		A list of available `pcolormesh` properties can be found here:
		https://matplotlib.org/api/_as_gen/matplotlib.pyplot.pcolormesh.html
	
	Returns
	-------
	None
	"""
	
	from numpy import arange, meshgrid
	from matplotlib.colors import LogNorm
	from matplotlib.pyplot import pcolormesh, colorbar
	from .base_func import axes_handler,plot_finalizer
	
	if ax is not None:
		old_axes=axes_handler(ax)
	if x is None:
		x=arange(len(im[:,0])+1)
	if y is None:
		y=arange(len(im[0,:])+1)
	if clog is None:
		from .defaults import Params
		clog=Params.img_caxis_log

	X, Y = meshgrid(x, y)

	# Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
	#plot_par = {**plot_kw, **kwargs} # For Python > 3.5
	plot_par = plot_kw.copy()
	plot_par.update(kwargs) 

	if clog:
		pcolormesh(X,Y,im.T,norm=LogNorm(vmin=clim[0],vmax=clim[1],clip=True),**plot_par)
	else:
		pcolormesh(X,Y,im.T,vmin=clim[0],vmax=clim[1],**plot_par)
	if clabel is not None:
		cbar=colorbar()
		cbar.set_label(clabel)
		if cbar_invert:
			cbar.ax.invert_yaxis()
	plot_finalizer(False,False,xlim,ylim,title,xlabel,ylabel,xinvert,yinvert,grid)
	if ax is not None:
		old_axes=axes_handler(old_axes)

# Scatter
def scatter(x,y,c=None,xlim=None,ylim=None,xinvert=False,yinvert=False,cbar_invert=False,xlog=False,ylog=False,title=None,
			xlabel=None,ylabel=None,clabel=None,plabel=None,lab_loc=0,ax=None,grid=None,plot_kw={},**kwargs):
	
	"""2D pixel-based image plotting function.
	
	Parameters
	----------
	x : array-like or list
		Position of data points in the x axis.
	y : array-like or list
		Position of data points in the y axis.
	c : array-like or list, optional
		Value of data points in the z axis (colour axis).
	xlim : tuple-like, optional
		Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
	ylim : tuple-like, optional
		Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
	xinvert : bool, optional
		If true inverts the x-axis.
	yinvert : bool, optional
		If true inverts the y-axis.
	cbar_invert : bool, optional
		If True inverts the direction of the colour bar (not the colour map).
	xlog : bool, optional
		If True the scale of the x-axis is logarithmic.
	ylog : bool, optional
		If True the scale of the x-axis is logarithmic.
	title : str, optional
		Sets the title of the plot
	xlabel : str, optional
		Sets the label of the x-axis.
	ylabel : str, optional
		Sets the label of the y-axis.
	clabel : str, optional
		Sets the legend for the colour axis.
	lab_loc : int, optional
		Defines the position of the legend
	ax : pyplot.Axes, optional
		Use the given axes to make the plot, defaults to the current axes.
	grid : boolean, optional
		If not given defaults to the value defined in splotch.Params.
	plot_kw : dict, optional
		Explicit dictionary of kwargs to be parsed to matplotlib scatter function.
		Parameters will be overwritten if also given implicitly as a **kwarg.
	**kwargs : Collection properties, optional
		kwargs are used to specify matplotlib specific properties such as cmap, marker, norm, etc.
		A list of available `Collection` properties can be found here:
		https://matplotlib.org/3.1.0/api/collections_api.html#matplotlib.collections.Collection
	
	Returns
	-------
	None
	"""

	from numpy import shape
	from matplotlib.pyplot import scatter, colorbar, legend
	from .base_func import axes_handler,dict_splicer,plot_finalizer
	
	if ax is not None:
		old_axes=axes_handler(ax)
	if type(x) is not list or len(shape(x))==1:
		x=[x]
	if type(y) is not list or len(shape(y))==1:
		y=[y]
	if type(c) is not list or len(shape(c))==1:
		c=[c]
	L=len(x)
	if type(plabel) is not list:
		plabel=[plabel]*L

	# Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
	#plot_par = {**plot_kw, **kwargs} # For Python > 3.5
	plot_par = plot_kw.copy()
	plot_par.update(kwargs)

	# Create 'L' number of plot kwarg dictionaries to parse into each scatter call
	plot_par=dict_splicer(plot_par,L,[len(i) for i in x])

	for i in range(L):
		scatter(x[i],y[i],c=c[i],label=plabel[i],**plot_par[i])
	if clabel is not None:
		cbar = colorbar()
		cbar.set_label(clabel)
		if cbar_invert:
			cbar.ax.invert_yaxis()
	if any(plabel):
		legend(loc=lab_loc)
	plot_finalizer(xlog,ylog,xlim,ylim,title,xlabel,ylabel,xinvert,yinvert,grid)
	if ax is not None:
		old_axes=axes_handler(old_axes)

def sector(r,theta,rlim=(0.0,1.0),thetalim=(0.0,360.0),rotate=0.0,rlabel="",thetalabel="",rstep=None,
			thetastep=15.0,rticks='auto',thetaticks='auto',fig=None,plot_kw={},**kwargs):
	
	""" Sector Plot function
	
	Plots a sector plot (a.k.a "pizza plot") based on data with one radial axis and an angular axis
	
	Parameters
	----------
	r : array-like or list
		Radial axis data.
	theta : array-like or list
		Angular axis data (degrees).
	rlim : tuple-like, optional
		The lower and upper limits for the radial axis (degrees).
	thetalim : tuple-like, optional
		The lower and upper limits for the angular axis (degrees).
	rotate : float, optional
		By how many degrees to rotate the entire plot (valid values in [-180, 180]).
	rlabel : str, optional
		Sets the label of the r-axis.
	thetalabel : str, optional
		Sets the label of the theta-axis.
	rstep : float, optional
		Sets the step size of r ticks.
	thetastep : float, optional, default: 15.0
		Sets the step size of theta ticks (degrees).
	rticks : 'auto', or ticker
		* Not implement *
	thetaticks : 'auto', or ticker
		* Not implement *
	fig : pyplot.Figure, optional
		Use the given figure to make the plot, defaults to the current figure.
	plot_kw : dict, optional
		Explicit dictionary of kwargs to be parsed to matplotlib scatter function.
		Parameters will be overwritten if also given implicitly in **kwargs.
	**kwargs : Collection properties, optional
		kwargs are used to specify matplotlib specific properties such as cmap, marker, norm, etc.
		A list of available `Collection` properties can be found here:
		https://matplotlib.org/3.1.0/api/collections_api.html#matplotlib.collections.Collection
	
	Returns
	-------
	ax : The pyplot.Axes object created for the sector plot.
	"""
	
	from matplotlib.transforms import Affine2D
	from matplotlib.projections.polar import PolarAxes
	from matplotlib.pyplot import gcf 
	
	from mpl_toolkits.axisartist import floating_axes
	from mpl_toolkits.axisartist.grid_finder import (FixedLocator, MaxNLocator, DictFormatter)
	import mpl_toolkits.axisartist.angle_helper as angle_helper
	
	from numpy import array, linspace, arange, shape, sqrt, floor, round, degrees, radians, pi
	
	if (fig == None):
		fig = gcf()
	
	# rotate a bit for better orientation
	trans_rotate = Affine2D().translate(0.0, 0)
	
	# scale degree to radians
	trans_scale = Affine2D().scale(pi/180.0, 1.)
	trans = trans_rotate + trans_scale + PolarAxes.PolarTransform()
	
	# Get theta ticks
	#if (thetaticks == 'auto'):
	thetaticks = arange(*radians(array(thetalim)+rotate),step=radians(thetastep))
	theta_gridloc = FixedLocator(thetaticks[thetaticks/(2*pi) < 1])
	theta_tickfmtr = DictFormatter(dict(zip(thetaticks,[f"{(round(degrees(tck)-rotate)):g}" for tck in thetaticks])))
	
	#tick_fmtr = DictFormatter(dict(angle_ticks))
	#tick_fmtr = angle_helper.Formatter()
	
	if (rstep == None):
		rstep = 0.5
	
	r_gridloc = FixedLocator(arange(rlim[0],rlim[1],step=rstep))
	
	grid = floating_axes.GridHelperCurveLinear(
		PolarAxes.PolarTransform(),
		extremes=(*radians(array(thetalim)+rotate), *rlim),
		grid_locator1=theta_gridloc,
		grid_locator2=r_gridloc,
		tick_formatter1=theta_tickfmtr,
		tick_formatter2=None,
	)
	
	ax = floating_axes.FloatingSubplot(fig, 111, grid_helper=grid)
	fig.add_subplot(ax)
	
	# tick
	thetadir_ref = ['top','left','bottom','right']
	rdir_ref = ['bottom','right','top','left']
	
	# adjust axis
	ax.axis["left"].set_axis_direction(rdir_ref[(int(rotate)//90)%4])
	ax.axis["right"].set_axis_direction(rdir_ref[(int(rotate)//90+2)%4])
	ax.axis["bottom"].set_axis_direction(thetadir_ref[(int(rotate)//90)%4])
	ax.axis["top"].set_axis_direction(thetadir_ref[(int(rotate)//90+2)%4])
	
	ax.axis["bottom"].set_visible(False if rlim[0] < (rlim[1]-rlim[0])/3 else True)
	ax.axis["bottom"].major_ticklabels.set_axis_direction(thetadir_ref[(int(rotate)//90 + 2)%4])
	
	ax.axis["top"].toggle(ticklabels=True, label=True)
	ax.axis["top"].major_ticklabels.set_axis_direction(thetadir_ref[(int(rotate)//90)%4])
	ax.axis["top"].label.set_axis_direction(thetadir_ref[(int(rotate)//90)%4])
	
	#ax.get_yaxis().set_major_locator(ticker.MaxNLocator())
	
	#ax.axis["left"].set_major_formatter(ticker.MaxNLocator())
	ax.axis["left"].major_ticklabels.set_axis_direction(rdir_ref[(int(rotate)//90)%4])
	ax.axis["left"].label.set_axis_direction(rdir_ref[(int(rotate)//90)%4])
	
	ax.axis["left"].label.set_text(rlabel)
	ax.axis["top"].label.set_text(thetalabel)
	
	# create a parasite axes whose transData in RA, cz
	sector_ax = ax.get_aux_axes(trans)
	
	# This has a side effect that the patch is drawn twice, and possibly over some other
	# artists. So, we decrease the zorder a bit to prevent this. 
	sector_ax.patch = ax.patch  
	sector_ax.patch.zorder = 0.9
	
	
	L = shape(theta)[0] if len(shape(theta)) > 1 else 1
	plot_par = plot_kw.copy()
	plot_par.update(kwargs)
	
	# Create 'L' number of plot kwarg dictionaries to parse into each plot call
	#plot_par = dict_splicer(plot_par,L,[1]*L)
	
	if (L == 1):
		sector_ax.scatter(theta+rotate, r, **plot_par)
	else:
		for ii in range(L):
			sector_ax.scatter(theta[ii]+rotate, r[ii],**plot_par[ii])
	
	return sector_ax

# Contours encircling the densest region down to a certain percentage 
def sigma_cont(x,y,percent=None,bin_type=None,bins=None,c=None,cmap=None,xlim=None,ylim=None,clim=None,xinvert=False,yinvert=False,
				cbar_invert=False,s=None,xlog=False,ylog=False,title=None,xlabel=None,ylabel=None,clabel=None,lab_loc=0,ax=None,
				grid=None,output=None,plot_kw={},**kwargs):
	
	"""Contour function, encircling the highest density regions that contain the given percentages of the sample (legacy name).
	
	This provides legacy compatibily for old code using the original name of the function.
	This function will eventually be removed, so consider switching to plots_2d.contourp().
	"""
	
	import warnings
	
	warnings.warn('This function provides legacy support and it will be removed in the future. Use plots_2d.contourp() instead.',FutureWarning)
	
	contourp(x,y,percent,bin_type,bins,c,cmap,xlim,ylim,clim,xinvert,yinvert,cbar_invert,s,xlog,ylog,title,xlabel,ylabel,
				clabel,lab_loc,ax,grid,output,plot_kw,**kwargs)

def contourp(x,y,percent=None,bin_type=None,bins=None,c=None,cmap=None,xlim=None,ylim=None,clim=None,xinvert=False,yinvert=False,
				cbar_invert=False,s=None,xlog=False,ylog=False,title=None,xlabel=None,ylabel=None,clabel=None,lab_loc=0,ax=None,
				grid=None,output=None,plot_kw={},**kwargs):
	
	"""Contour function, encircling the highest density regions that contain the given percentages of the sample.
	
	Parameters
	----------
	x : array-like
		Position of data points in the x axis.
	y : array-like
		Position of data points in the y axis.
	percent : float or list, optional.
		The percentages of the sample that the contours encircle.
	bin_type : {'number','width','edges','equal'}, optional
		Defines how is understood the value given in bins: 'number' for givinf the desired number of bins, 'width' for
		the width of the bins, 'edges' for the edges of bins, and 'equal' for making bins with equal number of elements
		(or as close as possible). If not given it is inferred from the data type of bins: 'number' if int, 'width' if
		float and 'edges' if ndarray.
	bins : int, float, array-like or list, optional
		Gives the values for the bins, according to bin_type.
	c : str, float or list, optional
		The colours of the contours. If float or list of floats, they must be in the range [0,1], as the colours are
		taken from the given colour map.
	cmap : str, optional
		The colour map to be used, viridis by default.
	xlim : tuple-like, optional
		Defines the limits of the x-axis, it must contain two elements (lower and higer limits).
	ylim : tuple-like, optional
		Defines the limits of the y-axis, it must contain two elements (lower and higer limits).
	clim : list, optional
		Defines the limits of the colour map ranges, it must contain two elements (lower and higer limits).
	xinvert : bool, optional
		If true inverts the x-axis.
	yinvert : bool, optional
		If true inverts the y-axis.
	cbar_invert : bool, optional
		If True inverts the direction of the colour bar (not the colour map).
	s= str or list, optional.
		Defines the linestyle of the contours.
	xlog : bool, optional
		If True the scale of the x-axis is logarithmic.
	ylog : bool, optional
		If True the scale of the x-axis is logarithmic.
	title : str, optional
		Sets the title of the plot
	xlabel : str, optional
		Sets the label of the x-axis.
	ylabel : str, optional
		Sets the label of the y-axis.
	clabel : str, optional
		Sets the legend for the colour axis.
	lab_loc : int, optional
		Defines the position of the legend
	ax : pyplot.Axes, optional
		Use the given axes to make the plot, defaults to the current axes.
	grid : boolean, optional
		If not given defaults to the value defined in splotch.Params.
	output : boolean, optional
		If True, returns the edges and values of the underlying histogram plus the levels of the contours.
	plot_kw : dict, optional
		Passes the given dictionary as a kwarg to the plotting function. Valid kwargs are QuadContourSet properties.
	**kwargs: QuadContourSet properties, optional
		kwargs are used to specify matplotlib specific properties such as cmap, linewidths, hatches, etc.
		The list of available properties can be found here: 
		https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.contour.html
	
	Returns
	-------
	bin_edges_x : array
		The bin edges for the x axis.
	bin_edges_y : array
		The bin edges for the y axis.
	n : array
		The values of the underlying histogram.
	l : array
		The levels for the contours.
	"""
	
	from matplotlib.cm import get_cmap
	from numpy import array,linspace, round
	from matplotlib.pyplot import gca, contour, legend
	from .base_func import axes_handler,basehist2D,percent_finder,plot_finalizer,dict_splicer
	
	if None in (percent,cmap,clim,s,output):
		from .defaults import Params
		if percent is None:
			percent=Params.sigcont_percent
		if cmap is None:
			cmap=Params.sigcont_cmap
		if clim is None:
			clim=Params.sigcont_clim
		if s is None:
			s=Params.sigcont_linestyle
		if output is None:
			output=Params.sigcont_output
	if ax is not None:
		old_axes=axes_handler(ax)
	if type(percent) is not list:
		percent=[percent]
	if type(bin_type) is not list:
		bin_type=[bin_type]*2
	if type(bins) is not list:
		if bins is None:
			bins=int((len(x))**0.4)
		bins=[bins]*2
	if output is None:
		from .defaults import Params
		output=Params.hist2D_output
	
	cmap=get_cmap(cmap)
	X,Y,Z=basehist2D(x,y,c,bin_type,bins,None,None,None,xlog,ylog)
	X=(X[:-1]+X[1:])/2
	Y=(Y[:-1]+Y[1:])/2
	CS=[]
	if c is None:
		if len(percent)<4:
			col_ax=gca()
			l=col_ax.plot([1,2,3])
			c=[l[0].get_color()]*len(percent)
			l.pop(0).remove()
		else:
			if len(s)<4:
				s=['solid']*len(percent)
			c=cmap(linspace(clim[0],clim[1],len(percent)))
	else:
		if type(c) is str:
			c=[c]*len(percent)
		else:
			c=cmap(c)
			s=['solid']*len(percent)
	if type(clabel) is not list:
		if clabel is None:
			clabel=[str(round(p,1))+'%' for p in percent]
		else:
			clabel= [clabel] + [None]*(len(percent)-1)
	if type(s) is not list:
		s=[s]*len(percent)
	
	# Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
	#plot_par = {**plot_kw, **kwargs} # For Python > 3.5
	plot_par = plot_kw.copy()
	plot_par.update(kwargs)
	
	# Create 'L' number of plot kwarg dictionaries to parse into each scatter call
	plot_par=dict_splicer(plot_par,len(percent),[1]*len(percent))
	
	level=[]
	for i in range(len(percent)):
		level.append(percent_finder(Z,percent[i]/100))
		CS.append(contour(X,Y,Z.T,levels=[level[i]],colors=[c[i],],linestyles=s[i],**plot_par[i]))
		if clabel[0] is not None:
			CS[i].collections[0].set_label(clabel[i])
	if clabel[0] is not None:
		legend(loc=lab_loc)
	plot_finalizer(xlog,ylog,xlim,ylim,title,xlabel,ylabel,xinvert,yinvert,grid)
	if ax is not None:
		old_axes=axes_handler(old_axes)
	if output:
		return(X,Y,Z.T,array(level))




def use_style(path):
	""" Loads in splotch configuration file containing a set of splotch style 
		parameters appended on top of a standard matplotlib .mplstyle configuration
		file.
	
	Parameters
	----------
	path : string
			The path to the configuration file.
	
	Returns
	-------
	None
	
	"""
	
	from .defaults import Params
	from .base_func import val_checker
	from matplotlib.pyplot import style
	
	(path)
	
	### SPLOTCH specific
	# Get list of available parameters within the splotch.Params class
	validParams = [key for key in Params.__dict__.keys() if key[:1] != '_']
	
	# find first Splotch line
	begin = None
	with open(path, 'r') as file:
		for num, line in enumerate(file.readlines()):
			if ("##### splotch configuration" in line.lower()):
				begin = num
				
	spltFile = open(path, 'r')
	allLines = spltFile.readlines()
	spltLines = allLines[begin:]
	pltLines = allLines[:begin]
	spltFile.close()
	pltDict ={}
	
	for line in spltLines:
		if (":" in line): # This line calls a parameter
			par = line.split(":")[0].rstrip()
			if par[0] == '#':
				continue
			else:
				val = line.split(":")[-1].lstrip().replace("\n","")
				
				# Adjust parameter value for booleans and numbers (i.e. floats/integer)
				if ',' in val:
					val = [val_checker(v) for v in val.split(',')]
				else:
					val = val_checker(val)
				
				if (par in validParams): # only edit parameters that exist within Params Class
					setattr(Params,par,val)
	
	for line in pltLines:
		if (":" in line): # This line calls a parameter
			par = line.split(":")[0].rstrip().lstrip()
			if par[0] == '#':
				continue
			else:
				val = line.split('#')[0].split(":")[-1].lstrip().replace("\n","").rstrip()
				
				# Adjust parameter value for booleans and numbers (i.e. floats/integer)
				if ',' in val and "'" not in val:
					val = [val_checker(v.rstrip()) for v in val.split(',')]
				else:
					val = val_checker(val.rstrip())
				
				pltDict[par]=val
	style.use(pltDict)
	
	return(None)

def reset_style():
	""" Resets the Splotch parameters to their defaults.
	
	Parameters
	----------
	None
	
	Returns
	-------
	None
	
	"""
	import os
	import splotch
	
	basedir = os.path.dirname(splotch.__file__)
	splotch.use_style("{0}/styles/default.style".format(basedir))
	
	return(None)




def adjust_text(which=['x','y'],ax=None,text_kw={},**kwargs):

	""" Adjusts text instances.

	Function which allows the user to adjust texts of one or many of either x/y-axis labels,
	title, legend, colorbars, etc.
	
	Parameters
	----------
	which : str, array-like or list
		Which `Text` instance(s) to apply the text properties to.

		Valid arguments can be one or many of the following:
			* 'x'|'xlabel'	 : x-axis label 
			* 'y'|'ylabel'	 : y-axis label 
			* 't'|'title'	 : Title
			* 's'|'suptitle' : Sup. title
			* 'l'|'legend'	 : Legend text
			* 'c'|'colorbar' : Color bar
			* 'T'|'text' 	 : Text objects
			* 'a'|'all'		 : All instances of all the above

	ax : pyplot.Axes or list, optional
		Use the given axes to adjust text, defaults to the current axis.
		If a list of Axis instances given, the text properties is applied to each.

	text_kw : dict, optional
		Explicit dictionary of kwargs to be parsed to matplotlib `Text` instance.
		It is recommended that text keyword arguments be given as **kwargs.

	**kwargs : Text instance properties
		kwargs are used to specify properties of `Text` instances
		A list of valid `Text` kwargs can be found here:
		[https://matplotlib.org/3.1.0/api/text_api.html#matplotlib.text.Text](https://matplotlib.org/3.1.0/api/text_api.html#matplotlib.text.Text "Matplotlib.text.Text")
	
	Returns
	-------
	None
	"""

	from matplotlib.text import Text
	from matplotlib.pyplot import gca
	from numpy import shape, min, max, argmin, argmax
	from .base_func import axes_handler,dict_splicer,plot_finalizer
	
	try:
		_ = (it for it in ax)
	except TypeError:
		if (ax == None):
			ax = gca()

		ax = [ax]

	# Validate `which` value(s)
	whichRef = ['x','y','t','s','l','c','T','a',
				'xlabel','ylabel','title','suptitle','legend','colorbar','text','all']

	try: # check if iterable
		_ = (i for i in which)
		if (type(which) == str):
			which = [which]

	except (TypeError):
		which = [which]

	for w in which:
		try:
			wInd = whichRef.index(w)
			wComp = (wInd + len(whichRef)//2) % len(whichRef) # get the complimenting short/long version
			if (whichRef[wComp] in which):
				raise TypeError("adjust_text() received equivalent values for 'which': '{0}' and '{1}'.".format(whichRef[wInd],whichRef[wComp]))

		except (ValueError):
			if (type(w) != Text):
				raise TypeError("adjust_text() received invalid value for 'which' ('{0}'). Must be one of: {1}".format(w,', '.join(whichRef)))
			

	L = len(which)
	
	# Combine the `explicit` plot_kw dictionary with the `implicit` **kwargs dictionary
	#plot_par = {**plot_kw, **kwargs} # For Python > 3.5
	textpar = text_kw.copy()
	textpar.update(kwargs)

	# Create 'L' number of plot kwarg dictionaries to parse into each plot call
	textpar = dict_splicer(textpar,L,[1]*L)

	for a in ax:
		for ii, lab in enumerate(which):
			if (lab in ['x','xlabel']):
				texts = [a.xaxis.label]
			elif (lab in ['y','ylabel']):
				texts = [a.yaxis.label]
			elif (lab in ['t','title']):
				texts = [a.title]
			elif (lab in ['s','suptitle']): # Not implemented
				texts = [a.title]
			elif (lab in ['l','legend']):
				texts = a.legend().get_texts() # Get list of text objects in legend
			elif (lab in ['c','colorbar']):
				# Get the axis with the largest ratio between width or height
				caxInd = argmax([max([c.get_position().width/c.get_position().height,c.get_position().height/c.get_position().width]) for c in a.figure.get_axes()])
				if (a.figure.get_axes()[caxInd].get_position().height > a.figure.get_axes()[caxInd].get_position().width):
					texts = [a.figure.get_axes()[caxInd].yaxis.label]
				else:
					texts = [a.figure.get_axes()[caxInd].xaxis.label]
			elif (lab in ['T','text']):
				texts = [child for child in a.get_children()[:-4] if type(child) == Text] # -4 to avoid grabbing Title, Subtitle. etc. Text instances
			elif (type(lab) == Text):
				texts = [lab]
			elif (lab in ['a','all']):
				texts = [a.xaxis.label,a.yaxis.label,a.title,a.legend().get_texts()]

				caxInd = argmax([max([c.get_position().width/c.get_position().height,c.get_position().height/c.get_position().width]) for c in a.figure.get_axes()])
				if (a.figure.get_axes()[caxInd].get_position().height > a.figure.get_axes()[caxInd].get_position().width):
					texts.append(a.figure.get_axes()[caxInd].yaxis.label)
				else:
					texts.append(a.figure.get_axes()[caxInd].xaxis.label)

				texts = texts + [child for child in a.get_children()[:-4] if type(child) == Text]

			for t in texts:
				t.set(**textpar[ii])
	
	return(None)


def subplots(naxes=None,nrows=None,ncols=None,va='top',ha='left',wspace=None,hspace=None,sharex='none',sharey='none',squeeze=True,figsize=None,axes_kw={},**kwargs):
	""" Adds a set of subplots to figure

	This is a more-generalised wrapper around matplotlib.pyplot.subplot function to allow for irregularly divided grids.
	
	Parameters
	----------
	naxes : int, optional, default: 1
		The number of axes objects to create.
		The resulting grid formed from specifying naxes is decided by ncols and nrows. The options are:

			- ncols and/or nrows not None:
				Makes sure that naxes can be correctly mapped into the specified grid.
				If one of nrows or ncols not given, the smallest possible grid will be made.
			- both ncols and nrows are None:
				Decides the best possible grid for this number of axes. Currently, this decision is hard-coded
				with plans for it to become an automatic decision later.
	
	nrows, ncols : int, optional
		Number of rows/columns of the subplot grid.
	
	va, ha : str, optional, default: 'top', 'left'
		The vertical alignment (va) and horizontal alignment (ha) sets the alignment of grids in the vertical and
		horizontal directions. 

		ha: 'left'		ha: 'centre'	ha: 'right'
		va: 'top'		va: 'top'		va: 'top'

		 ▯ ▯ ▯ ▯		 ▯ ▯ ▯ ▯		 ▯ ▯ ▯ ▯
		 ▯ ▯ ▯ ▯		 ▯ ▯ ▯ ▯		 ▯ ▯ ▯ ▯
		 ▯ ▯ ▯			  ▯ ▯ ▯			   ▯ ▯ ▯
		 ▯ ▯ ▯			  ▯ ▯ ▯			   ▯ ▯ ▯

		ha: 'left'		ha: 'centre'	ha: 'right'
		va: 'centre'	va: 'centre'	va: 'centre'

		 ▯ ▯ ▯							   ▯ ▯ ▯
		 ▯ ▯ ▯ ▯		   Not			 ▯ ▯ ▯ ▯
		 ▯ ▯ ▯ ▯		  Valid			 ▯ ▯ ▯ ▯
		 ▯ ▯ ▯							   ▯ ▯ ▯

		ha: 'left'		ha: 'centre'	ha: 'right'
		va: 'bottom'	va: 'bottom'	va: 'bottom'

		 ▯ ▯ ▯			  ▯ ▯ ▯			   ▯ ▯ ▯
		 ▯ ▯ ▯			  ▯ ▯ ▯			   ▯ ▯ ▯
		 ▯ ▯ ▯ ▯		 ▯ ▯ ▯ ▯		 ▯ ▯ ▯ ▯
		 ▯ ▯ ▯ ▯		 ▯ ▯ ▯ ▯		 ▯ ▯ ▯ ▯

	sharex, sharey : bool or {'none', 'all', 'row', 'col'}, default: False
		Not implemented.

	wspace : float, optional
		The horzontal spacing between figure subplots, expressed as a fraction of the subplot width.

	hspace : float, optional
		The vertical spacing between figure subplots, expressed as a fraction of the subplot height.

	sharex, sharey : bool or {'none', 'all', 'row', 'col'}, default: False
		As per matplotlib's usage, controls sharing of properties among x (`sharex`) or y (`sharey`)
		axes:

			- True or 'all': x-/y-axis will be shared among all subplots.
			- False or 'none': each subplot x-/y-axis will be independent (default).
			- 'row': each subplot row will share an x-/y-axis
			- 'col': each subplot column will share an x-/y-axis

		When subplots have a shared x-axis along a column, only the x tick
		labels of the last complete row of the subplot are created. Similarly, 
		when subplots have a shared y-axis along a row, only the y tick labels of the
		first complete column subplot are created. To later turn other subplots'
		ticklabels on, use `~matplotlib.axes.Axes.tick_params`.

	squeeze : bool, optional, default: True
		As per matplotlib's usage, the following applies:
			- If True, extra dimensions are squeezed out from the returned
			  array of Axes:

				- if only one subplot is constructed (nrows=ncols=1), the
				  resulting single Axes object is returned as a scalar.
				- for Nx1 or 1xM subplots, the returned object is a 1D numpy
				  object array of Axes objects.
				- for NxM, subplots with N>1 and M>1 are returned
				  as a 2D array.

			- If False, no squeezing at all is done: the returned Axes object
			  is always a 2D array containing Axes instances, even if it ends
			  up being 1x1.

		If naxes < ncols*nrows, the only sensible option is to return a 1D numpy array
	
	figsize : 2-tuple of floats, default: rcParams["figure.figsize"] * (ncols, nrows)
		The dimensions of the figure (width, height) in inches. If not specified, the default is to scale
		the default rcParams figure.figsize by the number of rows or columns.

	axes_kw : dict, optional
		Explicit dictionary of kwargs to be parsed to matplotlib `subplot` function.

	**kwargs : Text instance properties
		kwargs are used to specify properties of `subplots` instances
		A list of valid `axis` kwargs can be found here:
		[https://matplotlib.org/api/axes_api.html#matplotlib.axes.Axes](https://matplotlib.org/api/axes_api.html#matplotlib.axes.Axes "Matplotlib.axes.Axes")
	
	Returns
	-------
	fig : pyplot.Figure
	
	axes : pyplot.axes.Axes object or array of Axes objects
		axes may either be a single Axes object or a numpy array if naxes > 1.
		The dimensions of this array are controlled by the squeeze keyword above.

	"""

	from .base_func import dict_splicer
	
	import warnings
	from matplotlib import rcParams
	from matplotlib.gridspec import GridSpec
	from matplotlib.pyplot import figure, subplot
	
	from numpy import ceil, array, reshape, empty

	gridRef = [[0,0], [1,1], [1,2], [1,3], [2,2], [2,3], [2,3], [2,4], [2,4], [3,3], [2,5], [3,4], [3,4], 
			   [4,4], [3,5], [3,5], [4,4], [3,6], [3,6], [4,5], [4,5], [3,7], [5,5], [5,5], [4,6], [5,5]]


	if (isinstance(sharex, bool)):
		sharex = "all" if sharex else "none"
	if (isinstance(sharey, bool)):
		sharey = "all" if sharey else "none"

	if (naxes == None): # No number of axes specified
		if (nrows==None): nrows=1
		if (ncols==None): ncols=1
		
		naxes = nrows*ncols
	else:
		if (ncols == None and nrows == None):
			if (naxes <= 25):
				nrows, ncols = gridRef[naxes] # Get best combination of rows x cols for number of axes
			else:
				raise NotImplementedError(f"The naxes parameter is currently not implemented for naxes > 25.")
		elif (ncols != None and nrows != None):
			if (ncols*nrows != naxes):
				raise ValueError(f"Invalid number of axes ({naxes}) given for number of rows ({nrows}) and columns ({ncols}).") 
		else:
			if (nrows != None):
				ncols = int(ceil(naxes/nrows))
			else:
				nrows = int(ceil(naxes/ncols))
				
	
	# How many axes away from filling the gridspec evenly and completely
	delta = (nrows*ncols) - naxes
	
	if (figsize==None): # Auto scale default figure size to num cols/rows.
		figsize = (rcParams["figure.figsize"][0]*ncols, rcParams["figure.figsize"][1]*nrows)
	
	fig = figure(figsize=figsize,**kwargs)
	gs = GridSpec(ncols=ncols*2, nrows=nrows*2, hspace=hspace, wspace=wspace)
	
	axes_kw = dict_splicer(axes_kw,naxes,[1]*naxes)
	
	# Specify the row/col origin
	row0 = (nrows-1)*2 if va=='bottom' else 0
	col0 = (ncols-1)*2 if ha=='right' else 0

	axes = empty(naxes, dtype=object) # Create the empty array to hold each axis
	for ii in range(naxes):
		row = 2*(nrows-(ii//ncols)-1) if va=='bottom' else 2*(ii//ncols) # current row
		col = 2*(ncols-(ii%ncols)-1) if ha=='right' else 2*(ii%ncols) # current column

		# Select which axes this axis needs to be shared with
		sharewith = {"none": None, "all": axes[0],
					 "row": axes[(ii//ncols)*ncols], "col": axes[ii%ncols]}
		
		axes_kw[ii]["sharex"] = sharewith[sharex]
		axes_kw[ii]["sharey"] = sharewith[sharey]
			
		if (row == (0 if va=='bottom' else (nrows-1)*2) and ha=='centre'):
			axes[ii] = subplot(gs[row:row+2,col+delta:col+delta+2],**axes_kw[ii])
		elif (col == (0 if ha=='right' else (ncols-1)*2) and va=='centre'):
			axes[ii] = subplot(gs[row+delta:row+delta+2,col:col+2],**axes_kw[ii])
		else:
			axes[ii] = subplot(gs[row:row+2,col:col+2],**axes_kw[ii])


	# turn off redundant tick labeling
	if sharex in ["col", "all"]:
		if (ha=='centre'):
			warnings.warn("Removing redundant shared xtick labels not possible when ha='centre'")
		else:
			# turn off all but the bottom row
			for ax in axes[ncols:] if va=='bottom' else axes[:naxes-ncols]:
				ax.xaxis.set_tick_params(which='both',
										 labelbottom=False, labeltop=False)
				ax.xaxis.offsetText.set_visible(False)

	if sharey in ["row", "all"]:
		if (va=='centre'):
			warnings.warn("Removing redundant shared ytick labels not possible when va='centre'")
		else:
			# turn off all but the leftmost column
			for ii, ax in enumerate(axes):
				if (ha=='left'):
					if (ii%ncols!=0):
						ax.yaxis.set_tick_params(which='both',
												 labelbottom=False, labeltop=False)
						ax.yaxis.offsetText.set_visible(False)
				elif (ha=='right'):
					if (ii%ncols+1!=ncols and ii!=naxes-1):
						ax.yaxis.set_tick_params(which='both',
												 labelbottom=False, labeltop=False)
						ax.yaxis.offsetText.set_visible(False)


	### Squeeze axes object
	if (squeeze == True):
		# Discarding unneeded dimensions that equal 1.  If we only have one
		# subplot, just return it instead of a 1-element array.
		return (fig, axes.item()) if axes.size == 1 else (fig, axes.squeeze())
	else:
		# Returned axis array will be always 2-d, even if nrows=ncols=1.
		if (naxes != ncols*nrows):
			warnings.warn("squeeze = False not possible when naxes < nrows*ncols.")
			return (fig, axes.squeeze())
		else:
			return (fig, reshape(axes, (nrows, ncols)))


def axes_handler(new_axis):
	"""New axis handler
	
	This function is a base-level function used by most other plotting functions to set the current Axes instance
	to ax and returns the old Axes instance to be later reverted.
	
	Parameters
	----------
	new_axis : Axes object
		The new Axes instance to be set as the current axis.
	
	Returns
	-------
	curr_axis : Axes object
		The previous Axes instance.
	"""
	from matplotlib.pyplot import gca,sca
	
	curr_axis=gca()
	sca(new_axis)
	return(curr_axis)

def basehist2D(x,y,c,bin_type,bin_num,norm,dens,cstat,xlog,ylog):
	"""2D histogram base calculation
	
	This function is a base-level function used by plots_2d.hist2D() and plots_2d.sigma_cont() to calculate the
	underlying histogram.
	
	Parameters
	----------
	x : ndarray
		Position of data points in the x axis.
	y : ndarray
		Position of data points in the y axis.
	c : ndarray
		If a valid argument is given in cstat, defines the value used for the binned statistics.
	bin_type : {'number','width','edges','equal'}
		Defines how is understood the value given in bins: 'number' for givinf the desired number of bins, 'width' for
		the width of the bins, 'edges' for the edges of bins, and 'equal' for making bins with equal number of elements
		(or as close as possible). If not given it is inferred from the data type of bins: 'number' if int, 'width' if
		float and 'edges' if ndarray.
	bin_num : int, float or array-like
		Gives the values for the bins, according to bin_type.
	norm : float
		Normalization of the counts.
	dens : bool
		If false the histogram returns raw counts.
	cstat : str or function
		Must be one of the valid str arguments for the statistics variable in scipy.stats.binned_statistic_2d
		('mean’, 'median’, 'count’, 'sum’, 'min’ or 'max’) or a function that takes a 1D array and outputs an integer
		 or float.
	xlog : bool
		If True the scale of the x-axis is logarithmic.
	ylog : bool
		If True the scale of the x-axis is logarithmic.
	
	Returns
	-------
	x_bins_plot : ndarray
		The bin edges on the x-axis.
	y_bins_plot : ndarray
		The bin edges on the y-axis.
	Z : ndarray
		The value of each bin.
	"""
	
	from numpy import histogram2d
	from scipy.stats import binned_statistic_2d
	
	x_temp,x_bins_hist,x_bins_plot=bin_axis(x,bin_type[0],bin_num[0],log=xlog)
	y_temp,y_bins_hist,y_bins_plot=bin_axis(y,bin_type[1],bin_num[1],log=ylog)
	if cstat:
		Z=binned_statistic_2d(x_temp,y_temp,c,statistic=cstat,bins=[x_bins_hist,y_bins_hist])[0]
	else:
		Z=histogram2d(x_temp,y_temp,bins=[x_bins_hist,y_bins_hist],density=dens)[0]
		if dens and norm:
			Z*=1.0*len(x)/norm
	return(x_bins_plot,y_bins_plot,Z)

def bin_axis(data,btype,bins,log=False,plot_centre=False):
	"""Bin construction for histograms
	
	This function is a base-level function used by all histogram-related functions to construct the bins.
	
	Parameters
	----------
	data : ndarray
		Data to be binned.
	bin_type : {'number','width','edges','equal'}
		Defines how is understood the value given in bins: 'number' for givinf the desired number of bins, 'width' for
		the width of the bins, 'edges' for the edges of bins, and 'equal' for making bins with equal number of elements
		(or as close as possible). If not given it is inferred from the data type of bins: 'number' if int, 'width' if
		float and 'edges' if ndarray.
	bins : int, float, array-like or list
		Gives the values for the bins, according to bin_type.
	log: bool, optional
		If True, the bins are constructed in logarithmic space and the logarithm of the data is returned.
	step: bool, optional
		If True, returns the edges of the bins, instead of the midpoint values.
	
	Returns
	-------
	data : ndarray
		The data for the histogram.
	hist_bins : ndarray
		The bin edges.
	plot_bins: ndarray
		The bin centers.
	
	"""
	from numpy import linspace,nanmax,nanmin
	
	def N(d,b):
		if nanmin(d)==nanmax(d):
			h=linspace(nanmin(d)-0.5,nanmax(d)+0.5,num=b)
		else:
			h=linspace(nanmin(d),nanmax(d),num=b+1)
		return(h)
	
	def W(d,b):
		from math import ceil
		if nanmin(d)==nanmax(d):
			h=array([nanmin(d)-b/2,nanmax(d)+b/2])
		else:
			L=ceil((nanmax(d)-nanmin(d))/b)
			h=nanmin(d)+linspace(0,L*b,num=L)
		return(h)
	
	def E(d,b):
		return(b)
	
	def Q(d,b):
		if nanmin(d)==nanmax(d):
			h=linspace(nanmin(d)-0.5,nanmax(d)+0.5,num=b)
		else:
			from math import ceil,floor
			from numpy import array,concatenate,cumsum,ones,sort
			d=sort(d)
			L=len(d)
			w_l=floor(L/b)
			w_h=ceil(L/b)
			n_l=b*ceil(L/b)-L
			n_h=b-n_l
			n=concatenate([ones(ceil(n_l/2))*w_l,ones(n_h)*w_h,ones(floor(n_l/2))*w_l]).astype('int32')
			h=array([nanmin(d)]+[(d[i-1]+d[i])/2 for i in cumsum(n)[:-1]]+[nanmax(d)])
		return(h)
	
	if log:
		from numpy import log10
		data=log10(data)
	if btype is None:
		from numpy import ndarray
		bdict={int:'number',float:'width',ndarray:'edges'}
		btype=bdict[type(bins)]
	bfunc={'number':N,'width':W,'edges':E,'equal':Q}
	hist_bins=bfunc[btype](data,bins)
	plot_bins=hist_bins*1.0
	if plot_centre:
		plot_bins=(plot_bins[:-1]+plot_bins[1:])/2
	if log:
		plot_bins=10**plot_bins
	return(data,hist_bins,plot_bins)

def dict_splicer(plot_dict,Ld,Lx):
	"""Dictionary constructor for plotting
	
	This function is a base-level function used by most other plotting functions to construct a list of dictionaries,
	each containing the passed arguments to the underlying plotting calls for each different dataset.
	
	Parameters
	----------
	plot_dict : dict
		Contains the parameters to be passed to the underlying plotting function.
	Ld : int
		Number of plots to be made.
	Lx : list
		Contains the lenght of the data array of each plot to be made.
	
	Returns
	-------
	dict_list : list
		List of dictionaries, one for each plot to be made.
	"""
	dict_list=[]
	dict_keys=plot_dict.keys()
	#if 'rasterized' not in dict_keys:
	#	plot_dict['rasterized']=True
	for i in range(Ld):
		temp_dict={}
		for k in dict_keys:
			try:
				temp=(i for i in plot_dict[k])
			except TypeError:
				temp_dict[k]=plot_dict[k]
			else:
				if type(plot_dict[k]) is str or len(plot_dict[k])==Lx[i]:
					temp_dict[k]=plot_dict[k]
				else:
					temp_dict[k]=plot_dict[k][i]
		dict_list.append(temp_dict)
	return(dict_list)

def percent_finder(data,p):
	"""Level finder for percentage contours
	
	This function is a base-level function used by plots_2d.sigma_cont to define the level which contains the requested
	percentage of the data points.
	
	Parameters
	----------
	data : ndarray
		Describes the n-dimensional position of each data point.
	p : float
		Fraction of the data points to be encircled by the contour.
	
	Returns
	-------
	min_value : float
		Level for the contour.
	"""
	from numpy import cumsum,ravel,sort
	from numpy import min as np_min
	from numpy import sum as np_sum
	
	data_sorted=sort(ravel(data))[::-1]
	data_fraction=cumsum(data_sorted)
	data_fraction/=np_sum(data_sorted)
	try:
		min_value=np_min(data_sorted[data_fraction<p])
	except ValueError:
		min_value=np_min(data_sorted)
	return(min_value)

def plot_finalizer(xlog,ylog,xlim,ylim,title,xlabel,ylabel,xinvert,yinvert,grid_control):
	"""New axis handler
	
	This function is a base-level function used by most other plotting functions to set the current Axes instance
	to ax and returns the old Axes instance to be later reverted.
	
	Parameters
	----------
	xlog : None or bool
		If True, the x-axis scale is set to 'log'.
	ylog : None or bool
		If True, the y-axis scale is set to 'log'.
	xlim : None or array-like
		If given, defines the low and high limits for the x-axis. The first two elements must be int and/or float.
	ylim : None or array-like
		If given, defines the low and high limits for the y-axis. The first two elements must be int and/or float.
	title : None or str
		If given, defines the title of the figure.
	xlabel : None or str
		If given, defines the label of the x-axis.
	ylabel : None or str
		If given, defines the label of the y-axis.
	xinvert : None or bool
		If True, ensures the x-axis is inverted. If False, ensures the x-axis is not inverted.
	yinvert : None or bool
		If True, ensures the y-axis is inverted. If False, ensures the y-axis is not inverted.
	grid_control : None or bool
		If True, ensures the grid is turned on. If False, ensures the grid is turned off.
	
	Returns
	-------
	None
	"""
	from .defaults import Params
	from matplotlib.pyplot import gca,grid,xscale,yscale
	from matplotlib.pyplot import title as plt_title, xlabel as plt_xlabel, xlim as plt_xlim, ylabel as plt_ylabel, ylim as plt_ylim
	
	if xlog:
		xscale('log')
	if ylog:
		yscale('log')
	if xlim is not None:
		plt_xlim(xlim)
	else:
		plt_xlim(auto=True)
	if ylim is not None:
		plt_ylim(ylim)
	else:
		plt_ylim(auto=True)
	if title is not None:
		plt_title(title)
	if xlabel is not None:
		plt_xlabel(xlabel)
	if ylabel is not None:
		plt_ylabel(ylabel)
	if xinvert:
		if not gca().xaxis_inverted():
			gca().invert_xaxis()
	if yinvert:
		if not gca().yaxis_inverted():
			gca().invert_yaxis()
	if grid_control is None:
		grid_control=Params.grid
	grid(b=grid_control,which=Params.grid_which,axis=Params.grid_axis)

def step_filler(x,y,**kwargs):
	"""Wrapper for a specific fill_between plot.
	
	Parameters
	----------
	x : array-like
		The x-axis values.
	y : array-like
		The y-axis values.
	Other arguments : **kwargs
		Optional kwargs supported by fill_between. Note that it will conflict if 'step' is given.
	Returns
	-------
	val : Recasted value
	"""
	
	from numpy import empty
	from matplotlib.pyplot import fill_between
	
	temp_y=empty(len(y)+1)
	temp_y[1:]=y
	temp_y[0]=y[0]
	fill_between(x,temp_y,step='pre',**kwargs)
	
	return(None)

def val_checker(val):
	"""Type check of variable to pass to style functions.
	
	Parameters
	----------
	val : str
		String containing the value to evaluate
	Returns
	-------
	val : Recasted value
	"""
	
	from distutils.util import strtobool
	
	if "'" not in val and '"' not in val:
		try: 
			val=float(val) if '.' in val else int(val)
		except ValueError:
			try:
				val=bool(strtobool(val))
			except ValueError:
				val=val
				#if val=='None':
				#	val=None
	
	return(val)



import atexit
from setuptools import setup
from setuptools.command.install import install

setup(name='splotch',
		version='0.4.0.0',
		description='Simple PLOTs, Contours and Histograms is a small package with wrapper functions designed to simplify plotting calls from matplotlib.',
		url='https://github.com/MBravoS/splot',
		author='Matías A. Bravo Santa Cruz',
		author_email='matias.bravo@icrar.org',
		license='BSD-3-Clause',
		packages=['splotch'],
		package_dir={'splotch': 'src/splotch'},
		include_package_data=True,
		package_data={'src':['styles/*.style']},
		#cmdclass={'install':new_install},
		install_requires=['numpy>=1.10','matplotlib>=2.2.0','scipy>=1.0.0'],
		zip_safe=False)

